////////////////////////////////////////////////////////////////////
// WMCA Markov Chain Analyser        (C) 1995 William Knottenbelt //
////////////////////////////////////////////////////////////////////
//            Data Network Architectures Laboratory               //
//               Department of Computer Science                   //
//                   University of Cape Town                      //
//                       Rondebosch 7700                          //
//                        SOUTH AFRICA                            //
////////////////////////////////////////////////////////////////////
// email: william@cs.uct.ac.za             phone: +27 21 650 2670 //
////////////////////////////////////////////////////////////////////

//#define BUFFER_SIZE 16384
//#define VERBOSE
//#define DEBUG
//#define SILENT

using namespace std;

#include <iostream>
#include <iomanip>
#include <cassert>
#include <cstdio>
#include <cstdarg>
#include <cstdlib>
#include "common.hpp"
#include "timer.hpp"
#include "dyna.hpp"
#include "bit.hpp"
#include "info.hpp"
#include "element.hpp"
#include "func.hpp"
#include "filenames.hpp"
#ifdef RESOURCE_USAGE
#include <unistd.h>
#endif

FunctionalAnalyser::FunctionalAnalyser(char* optionsFileName, char* infoFileName) {
#ifdef LOGGING
  cout << "FunctionalAnalyser (" << this << "): powering up..." << endl;
#endif
#ifdef RESOURCE_USAGE
  core = sbrk(0);
#endif
  states = 0;
  trans = 0;
  components = 0;
  height = 0;
  info = new Information;
  timer = new Timer;
  timer->start();

  if (!info->input(optionsFileName)) 
     terminate("func.cxx cannot read options file");
  if (!info->inputStateInfo(infoFileName)) 
     terminate("func.cxx cannot read state information file");

  repInterval = info->getGenRepInterval();
  repStyle = info->getGenRepStyle();
  transitions = info->getTransitions();
  if (!info->getGeneration()) terminate ("run state space generator first");
  if (!info->getTangible()) terminate("no tangible states");
  #ifdef LOGGING
  cout << "FunctionalAnalyser (" << this << "): powered up!" << endl;
  #endif
}

FunctionalAnalyser::~FunctionalAnalyser() {
  destroy();
}

void FunctionalAnalyser::destroy() {
  #ifdef LOGGING
  cout << "FunctionalAnalyser (" << this << "): destructor..." << endl;
  #endif
  for (long n=0; n<states; n++)
    delete state[n];
  delete info;
  delete timer;
  #ifdef LOGGING
  cout << "FunctionalAnalyser (" << this << "): powered down..." << endl;
  #endif
}

void FunctionalAnalyser::terminate(char *fmt, ...)
{
  char buffer[2048];

  va_list argp;
  va_start(argp, fmt);
  vsprintf(buffer, fmt, argp); 
  va_end(argp);
  destroy();
  printf("\nfatal error: %s.\nexecution halted.\n", buffer);  
  exit(1);
}

void FunctionalAnalyser::findStronglyConnectedComponents(char* infoFileName)
{
  #ifdef LOGGING
  cout << "FunctionalAnalyser (" << this << "): finding strongly connected components..." << endl;
  #endif

  timer->read();
  //  cout << "0!!!" << endl;
  //  exit(1);
  SCompInitialise();  
  //  cout << "1!!!" << endl;
  //  if (states) SCompDFS(0);
  for(long n=0; n<states; n++) {
#ifdef DEBUG
    assert(n>=0 && n<states);
#endif
    if (!dfsNumber(n)) SCompDFS(n);
  }
  //  cout << "2!!!" << endl;
  SCompAnalyse(infoFileName);
  //  cout << "3!!!" << endl;
  SCompFinish();
  //  cout << "4!!!" << endl;
  #ifdef LOGGING
     double timeUsed = timer->read();
     cout << "FunctionalAnalyser (" << this << "): components analysed (" << (double) timeUsed << " CPU seconds)" << endl;
  #endif
//  statistics();
}

void FunctionalAnalyser::SCompInitialise()
{
  #ifdef DEBUG
  cout << "states = " << states << endl;
  #endif
  //  cout << "1.1!!" << endl;
#ifdef DEBUG
  for (long n=0; n<states; n++)
    assert(state[n]->magic == 1973);
  cout << "passed magic test 2!" << 20.20 << endl;
#endif
  dfsNumber.setLength(states);
  //  cout << "1.2!!" << endl;
  low.setLength(states);
  //  cout << "1.3!!" << endl;
  removed.setLength(states);
  //  cout << "1.4!!" << endl;
  removed.clear();
  //  cout << "1.5!!" << endl;

  stack.setLength(states);
  //  cout << "1.6!!" << endl;
  local.setLength(1024);
  //  cout << "1.7!!" << endl;

  dfn = 0; 
  for (long n=0; n<states; n++) {
    dfsNumber[n] = 0;
    low[n] = n;
  }
  #ifdef DEBUG
  cout << "end scomp initialise" << endl;
  #endif
}

void FunctionalAnalyser::SCompDFS(long v)
{
  int t;
  long w, popped;
  Node *current;

  level = 0;

start:
  dfn++;
  dfsNumber[v] = dfn;
  low[v] = dfn;
//  removed.set0(v);
//  removed[v] = 0;

  current = state(v);
  for (t=0; t<current->numberEnabled(); t++) {
    w = current->getChild(t); //.getDestination();
//      cout << "Destination: " << w << "\n";
    if (!dfsNumber(w)) {
      pushLocalVars(v,w,t);
      v = w;
      current = state(v);
      goto start;
resume:
      popLocalVars(v,w,t);
//      if (v<0 || v>=states) cout << "BLUE MURDER! " << v << "\n";
      current = state(v);
//      SCompDFS(w);
      // now backing up from w to v
      if (low(w) < low(v)) low[v] = low(w);
    } else { // w was already encountered
      if (!removed(w) && dfsNumber(w) < low(v)) low[v] = dfsNumber[w];
    }
  } // end for

  // now backing up from v
  if(low[v]==dfsNumber[v]) {
    members.check(components);
    members[components] = 1;
//    cout << "new strongly connected component found!\n";
    removed.set1(v);
    state[v]->setComponent(components);
    while (height && dfsNumber[top()] > dfsNumber[v]) {
      popped = pop();
      removed.set1(popped);
      state[popped]->setComponent(components);
      members[components]++;
    }
    components++;
  } else {
    push(v);
  }
  if (level) goto resume;
}

void FunctionalAnalyser::SCompFinish()
{
//  cout << "finish started\n";
  stack.setLength(0);
  local.setLength(0);
  dfsNumber.setLength(0);
  low.setLength(0);
  removed.setLength(0);
//  cout << "finish complete\n";
}

void FunctionalAnalyser::SCompAnalyse(char* infoFileName)
{
//  cout << "Analyse begin! ( " << components << " components, " << transitions << " transitions)\n";

  final.setLength(components);
  transient.setLength(states);
  transient.clear();

  long c,n;
  for (c=0; c<components; c++)
    final.set1(c);

  long currentComponent; 
  int t, successors;
  Node *current;
  for (n=0; n<states; n++) {
    current = state(n);
    currentComponent = current->getComponent();
    #ifdef SAFE
    if(currentComponent == _UNDEFINED_)
      terminate("internal error (undefined component)");
    #endif
    successors = current->numberEnabled();
    for (t=0; t < successors; t++) {
      if (state(current->getChild(t))->getComponent() != currentComponent) {
#ifdef DEBUG
        cout << "Component " << currentComponent << " is not final!!\n"; // ###
	//	cout << "destination " << current->getChild(t).getDestination() << " is in component " << state(current->getChild(t).getDestination())->getComponent() << "!=" << currentComponent << "\n";
#endif
        final.set0(currentComponent);
      }
    }
  }

  // establish number of final strongly connected components
  int numberFinal = 0;

  for (c=0; c<components; c++)
    if (final(c)) numberFinal++;

  transients=0;
  for (n=0; n<states; n++) 
    if (!final(state(n)->getComponent())) {
      transients++;
      transient.set1(n);
    }

  #ifdef LOGGING
  double timeUsed = timer->read();
  cout << "FunctionalAnalyser (" << this << "): connected components analysed (" << timeUsed << " seconds):" << endl << endl;
  cout << states << " states and " << trans << " transitions." << endl;
  cout << components << " state class(es) (" << components - numberFinal << " transient, " << numberFinal << " recurrent)\n" << endl;
  if (repStyle != NONE) {
    for (c=0; c<components; c++) {
      cout << "State class " << c << " (" << members(c) << " states) is ";
      if (final(c)) cout << "recurrent."; else cout << "transient.";
      cout << endl;
    }
  }
  cout << endl;
  if (repStyle == FULL) {
    cout << maxHeight << " was the maximum DFS stack height.\n";
    cout << maxLevel << " was the maximum component stack height.\n";
    cout << endl;
  }
  #endif
  if (numberFinal == 0)
    terminate("no recurrent classes");
  if (numberFinal > 1)
    info->setReducible(1);

  if (transients) remap();

  info->setFunctional(1);
#ifdef RESOURCE_USAGE
  unsigned long memory = (unsigned long) (sbrk(0) - core);
  cout << "FunctionalAnalyser (" << this << "): " << (double) (memory/1024.0) << " KB memory used for data." << endl;
#endif
  if (!info->outputStateInfo(infoFileName))
    terminate("cannot update state information file");
}

int FunctionalAnalyser::readStates() {
  char *filename = info->getTransMatrixFile();
  #ifdef LOGGING
  cout << "FunctionalAnalyser (" << this << "): reading states..." << endl;
  #endif
  timer->read();
  state.setLength(info->getTangible());
  #ifdef LOGGING
  cout << info->getTangible() << " tangible states (max. children = " <<
    info->getMaxChildren() << ")" << endl;
  #endif
  char matrixBuffer[BUFFER_SIZE];
  FILE *mf = fopen(filename,"rb");
  if (!mf) terminate("cannot open state matrix file '%s'", filename);
  if (setvbuf(mf, matrixBuffer, _IOFBF, BUFFER_SIZE))
      terminate("cannot allocate buffer for '%s'", filename);

  int kids, maxKids = info->getMaxChildren();
#ifdef SAFE
  long maxStates = info->getTangible();
#endif

  StateMatrixElement *child = new StateMatrixElement[maxKids];
  //StateMatrixElement *child = (StateMatrixElement *) new char[sizeof(StateMatrixElement)*maxKids];
#ifdef SAFE
  assert(child);
#endif

  while(fread(&kids, sizeof(int), 1, mf)) {
//    child = new StateMatrixElement[kids];
        // cout << "state: " << states << " (" << kids << " kids) " << endl;
#ifdef SAFE
    assert(kids > 0);
    assert(kids <= maxKids);
    assert(states < maxStates);
#endif
    fread(child, sizeof(StateMatrixElement), kids, mf);

    //     for (int n=0; n<kids; n++)
    //       cout << n << " destination " << child[n].getDestination() << " weight " << child[n].getWeight() << endl;
    
//    state.check(states);
//    assert(child);
    state[states] = new Node(kids, child);
#ifdef DEBUG
    assert(state[states]);
    assert(state[states]->magic == 1973);
#endif
    //    cout << "State " << states << " with " << kids << " children\n";
    trans += kids;
    states++;
//    cout << "current state: " << states << endl;
    #ifdef LOGGING
    if (repInterval && !(states % repInterval) && repStyle != NONE) cout << "Reading state " << states << "..." << endl;
    #endif
//    delete child;
  }
  //  cout << "deleting child!" << 20.18 << endl;
  delete [] child;
  //  cout << "closing file!" << 20.19 << endl;
  fclose(mf);
  assert(states == info->getTangible());
#ifdef DEBUG
  assert(timer);
  for (long n=0; n<states; n++)
    assert(state[n]->magic == 1973);
  cout << "passed magic test!" << 20.20 << endl;
#endif
  #ifdef LOGGING
  double timeUsed = timer->read();
  cout << "FunctionalAnalyser (" << this << "): read complete ( " << states << " states, " << trans << " transitions)" << endl;
  if (timeUsed) {
    cout << "(read speed " << ((double) states/timeUsed) << " states/second, " << ((double) trans/timeUsed) << " transitions/second)" << endl;
  }
  #endif
  return 1;
}

Node::Node(int kids, StateMatrixElement *child)
{
  double weight;
#ifdef DEBUG
  magic = 1973;
#endif    
  offspring = 0;
  children.setLength(kids);
  component = _UNDEFINED_;
  for (int n=0; n < kids; n++) {
    weight = child[n].getWeight();
    if (weight > 0) {
      children[offspring] = child[n].getDestination();
      offspring++;
    } 
  } 
  if (offspring != kids) children.setLength(offspring);
#ifdef DEBUG
  if (offspring != kids) cout << "WARNING: offspring != kids" << endl;
  assert(magic==1973);
#endif
}

int FunctionalAnalyser::remap() {

#ifdef LOGGING
  cout << "FunctionalAnalyser (" << this << "): remapping..." << endl;
#endif
  timer->read();
  long n;
  recurrent = 0;
  mapping.setLength(states);
  for (n=0; n<states; n++) {
    if (transient(n))
      mapping[n] = -1;
    else {
      mapping[n] = recurrent;
      recurrent++;
    }
  }
  if (!remapMatrix()) terminate("transition matrix remapping failed");
  if (!remapStates()) terminate("state remapping failed");
  info->setTangible(recurrent);
  mapping.setLength(0);

  #ifdef LOGGING
  double timeUsed = timer->read();
  cout << "FunctionalAnalyser (" << this << "): remap complete (" << states << " states -> " << recurrent << " states)" << endl;
  if (timeUsed)
    cout << "(remap speed " << ((double) states/timeUsed) << " states/second)" << endl;
  #endif
  return 1;
}

int FunctionalAnalyser::remapMatrix() {
  char *filename = info->getTransMatrixFile();
  #ifdef LOGGING
  cout << "FunctionalAnalyser (" << this << "): remapping transition matrix..." << endl;
  #endif
  char matrixBuffer[BUFFER_SIZE], outputBuffer[BUFFER_SIZE];
  FILE *mf = fopen(filename,"rb");
  if (!mf) terminate("cannot open state matrix file '%s'", filename);
  if (setvbuf(mf, matrixBuffer, _IOFBF, BUFFER_SIZE))
      terminate("cannot allocate buffer for '%s'", filename);
  FILE *rf = fopen("REMAP", "wb");
  if (!rf) terminate("cannot open state matrix remap file 'REMAP'");
  if (setvbuf(rf, outputBuffer, _IOFBF, BUFFER_SIZE))
      terminate("cannot allocate buffer for 'REMAP'");

  int kids, maxKids = info->getMaxChildren();
#ifdef SAFE
  long maxStates = info->getTangible();
#endif

  StateMatrixElement *child = new StateMatrixElement[maxKids];
#ifdef SAFE
  assert(child);
#endif
  states = 0;
  recurrent = 0;
  while(fread(&kids, sizeof(int), 1, mf)) {
#ifdef SAFE
    assert(kids > 0);
    assert(kids <= maxKids);
    assert(states < maxStates);
#endif
    fread(child, sizeof(StateMatrixElement), kids, mf);

    if (!transient(states)) {
      recurrent++;
      for (int n=0; n<kids; n++) {
	long dest = child[n].getDestination();
	assert(mapping[dest]!=-1);
	child[n].setDestination(mapping(dest));
      }
      fwrite(&kids, sizeof(int), 1, rf);
      fwrite(child, sizeof(StateMatrixElement), kids, rf);
    }
    trans += kids;
    states++;
    #ifdef LOGGING
    if (repInterval && !(states % repInterval) && repStyle != NONE) cout << "Remapping state " << states << "..." << endl;
    #endif
//    delete child;
  }
  delete [] child;
  fclose(mf);
  fclose(rf);
  assert(states == info->getTangible());
  if (rename("REMAP", filename))
    terminate("could not rename 'REMAP' -> '%s'", filename);
  return 1;
}

int FunctionalAnalyser::remapStates() {
#ifdef LOGGING
  cout << "FunctionalAnalyser (" << this << "): remapping state space..." << endl;
#endif
  //long n ;
  long size = info->getStateSize();

  char stateBuffer[BUFFER_SIZE], outputBuffer[BUFFER_SIZE];
  char *filename = info->getStateListFile();
  FILE *sf = fopen(filename,"rb");
  if (!sf) terminate ("cannot open state file '%s'", filename);
  if (setvbuf(sf, stateBuffer, _IOFBF, BUFFER_SIZE))
    terminate("cannot allocate buffer for '%s'", filename);
  FILE *rf = fopen("REMAP", "wb");
  if (!rf) terminate("cannot open state matrix remap file 'REMAP'");
  if (setvbuf(rf, outputBuffer, _IOFBF, BUFFER_SIZE))
      terminate("cannot allocate buffer for 'REMAP'");

  char *current = new char[size];
  states=0;
  recurrent=0;
  while(fread(current, size, 1, sf)) {
    if (!transient(states)) {
      recurrent++;
      fwrite(current, size, 1, rf);
    }
    states++;
  }
  delete [] current;
  fclose(sf);
  fclose(rf);
  assert(states == info->getTangible());
  if (rename("REMAP", filename))
    terminate("could not rename 'REMAP' -> '%s'", filename);
  return 1;
}


int main(int argc, char *argv[])
{
   if (argc != 2)
   {
      printf("Must supply the options file as an argument\n") ;
      printf("Number of arguments was: %d\n", argc) ;
      exit(1) ;
   }

   char* basename        = argv[1] ;
   char* optionsFileName = add_extension(basename, "OPTIONS") ;
   char* infoFileName    = add_extension(basename, "INFO") ;

  FunctionalAnalyser *func = new FunctionalAnalyser(optionsFileName, infoFileName) ;

  func->readStates();  
  func->findStronglyConnectedComponents(infoFileName);
  delete func;
  return 0;
}

