
//  Program to launch the various parts of HYDRA-S in sequence
//  by Nick Dingle <njd200@doc.ic.ac.uk> 08/10/02
//  after "driver.cxx" by Will Knottenbelt


using namespace std;

#include <iostream>
#include <iomanip>
#include <cstdio>
#include <cstring>
#include <cstdlib>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include "common.hpp"
#include "filenames.hpp"


int fail() {
  return -1;
}

/* #define LOGGING 1 */

int execute(char *args[], int parameters) {
#ifdef LOGGING
  for (int n=0; n<parameters; n++)
    cout << args[n] << " ";
  cout << endl;
#endif

  args[parameters] = NULL;
  long pid = fork();
  int status;
  if (pid==0) {
    execvp(args[0], args);
    cout << "execution failure (pid == 0)." << endl;
    exit(-1);
  }
  if (pid < 0) {
    cout << "execution failure (pid < 0)."  << endl;
    exit(-1);
  }
  while(1) {
    if (waitpid(pid, &status, 0)) {
      if (status != 0) 
	cout << "return code is " << status << endl;
      return !status;
    }
  }
}

main(int argc, char *argv[]) {
  char *args[20];
  char *install = getenv("HYDRA_S_DIR");
  char libraries[512], headers[512];
  char usrdir[MAX_LENGTH]; 
  char libdir[MAX_LENGTH], hppdir[MAX_LENGTH];

  if (!install) {
    cout << argv[0] << ": set environment variable HYDRA_S_DIR to HYDRA-S's install directory" << endl;
    return 1;
  }

  if (argc < 2) {
    cout << argv[0] << ": no input file specified" << endl;
    return 1;
  }

  sprintf(libdir, "%s/hydra/lib", install);
  sprintf(hppdir, "%s/hydra/hpp", install);

 
  /* Make the names for all the produced files here */
  char* filename = argv[1] ;

  char* basenamestring    = remove_extension(filename) ;

  char* userCxxname       = add_extension(basenamestring, "user.cxx") ;
  char* userObjectFile    = add_extension(basenamestring, "user.o") ;

  /*
   This assumes you're using a relative path name.
  */
  char* pathPlusBase      = add_relative_start(basenamestring) ;
  char* genProgram        = add_extension(pathPlusBase, "gen" ) ;
  char* gspnTargetProgram = add_extension(pathPlusBase, "gspn-target" ) ;
  char* perfProgram        = add_extension(pathPlusBase, "perf" ) ;


  // parse <name>
  // make -f gspn-statemake
  // gen
  // make -f gspn-steadymake
  // gspn-steady
  // make -f gspn-targetmake
  // gspn-target
  // make -f gspn-passagemake


#ifdef LOGGING
  cout << endl << "Parsing input file..." << endl;
#endif
  sprintf(usrdir, "%s/hydra-parse", install);
  //  args[0] = "parse";
  args[0] = usrdir;
  args[1] = filename ;
  args[2] = basenamestring ;
  if (!execute(args, 3)) return fail();

#ifdef LOGGING
  cout << endl;

  cout << endl << "Compiling user code..." << endl;
#endif
  args[0] = "g++";
  //  args[1] = "-g";
  args[1] = "-O2";
  args[2] = "-o" ;
  args[3] = userObjectFile ;
  args[4] = "-c";
  sprintf(headers,"-I%s",hppdir);
  args[5] = headers;
  args[6] = userCxxname ;
  if (!execute(args, 7)) return fail();

#ifdef LOGGING
  cout << endl << "Linking state generator..." << endl;
#endif
  args[0] = "g++";
//  args[1] = "-g";
  args[1] = "-O2";
  args[2] = userObjectFile ;
  sprintf(libraries,"-L%s",libdir);
  args[3] = libraries;
  args[4] = "-o";
  //  sprintf(usrdir, "%s/gen", install);
  //  args[5] = usrdir;
  args[5] = genProgram;
  args[6] = "-lstate";
  args[7] = "-lcommon";
  args[8] = "-lm";
  args[9] = "-lstdc++";
  if(!execute(args, 10)) return fail();

#ifdef LOGGING
  cout << endl << "Generating state space..." << endl;
#endif
  //  sprintf(usrdir, "%s/gen", install);
  //  args[0] = usrdir;
  args[0] = genProgram ;
  args[1] = basenamestring ;
  if (!execute(args, 2)) return fail();

#ifdef LOGGING
  cout << endl << "Performing functional analysis..." << endl;
#endif
  sprintf(usrdir, "%s/hydra-func", install);
  args[0] = usrdir;
  args[1] = basenamestring ;
  if (!execute(args, 2)) return fail();

#ifdef LOGGING
  cout << endl << "Calling steady state analyser..." << endl;
#endif
  sprintf(usrdir, "%s/hydra-steady", install);
  args[0] = usrdir;
  args[1] = basenamestring ;
  if (!execute(args, 2)) return fail();

#ifdef LOGGING
  cout << endl << "Linking target state identifier..." << endl;
#endif
  args[0] = "g++";
  args[1] = "-O";
  args[2] = userObjectFile ;
  sprintf(libraries,"-L%s",libdir);
  args[3] = libraries;
  args[4] = "-o";
  //  sprintf(usrdir, "%s/gspn-target", install);
  //  args[5] = usrdir;
  args[5] = gspnTargetProgram ;
  args[6] = "-ltarget";
  args[7] = "-lcommon";
  args[8] = "-lm";
  args[9] = "-lstdc++";
  if(!execute(args, 10)) return fail();

#ifdef LOGGING
  cout << endl;

  cout << "Identifying target states..." << endl;
#endif
  //sprintf(usrdir, "%s/gspn-target", install);
  //args[0] = usrdir;
  args[0] = gspnTargetProgram ;
  args[1] = basenamestring ;
  if (!execute(args, 2)) return fail();

#ifdef LOGGING
  cout << endl;

  cout << "Linking performance analyser..." << endl;
#endif
  args[0] = "g++";
  args[1] = "-O";
  args[2] = userObjectFile ;
  sprintf(libraries,"-L%s",libdir);
  args[3] = libraries;
  args[4] = "-o";
  //  sprintf(usrdir, "%s/perf", install);
  //  args[5] = usrdir;
  args[5] = perfProgram ;
  args[6] = "-lperf";
  args[7] = "-lcommon";
  args[8] = "-lm";
  args[9] = "-lstdc++";
  if(!execute(args, 10)) return fail();

  cout << endl;

  //sprintf(usrdir, "%s/perf", install) ;

  //  cout << "For steady-state analysis, execute \"perf\"" << endl;
  cout << "For steady-state analysis, execute \"" << perfProgram << " "
       << basenamestring << "\"" << endl;
  cout << "  -- results aren't written to disk, so use \"tee\" or pipe to a file to save them" << endl;

  cout << endl;

  sprintf(usrdir, "%s/hydra-uniform", install);

  //  cout << "For response-time/transient analysis, execute \"uniform\"" << endl;
  cout << "For response-time/transient analysis, execute \"" << usrdir << 
       " " << basenamestring << "\"" << endl;
  cout << "  -- use the \"-cdf\" option to produce the distribution rather than the density" << endl;
  cout << "  -- in either case there is no need to specify the model filename" << endl;
  cout << "  -- results are written to the file \"PT_RESULTS\" as <t-point> <value>" << endl;
  cout << "  -- or you can grep the output for \"DATA0\" to see the results as well" << endl;

  /*
  cout << "Compiling response-time analyser..." << endl;
  args[0] = "make";
  args[1] = "-f";
  sprintf(usrdir, "%s/uniformmake", install);
  args[2] = usrdir;
  if (!execute(args, 3)) return fail();
  */
  /*
  sprintf(usrdir, "%s/uniform", install);
  args[0] = usrdir;
  if (!execute(args, 1)) return fail();
  */

  cout << endl;
 
  cout << "Execution complete." << endl << endl; 

  return 0;
}





