////////////////////////////////////////////////////////////////////
// WMCA Markov Chain Analyser        (C) 1995 William Knottenbelt //
////////////////////////////////////////////////////////////////////
//            Data Network Architectures Laboratory               //
//               Department of Computer Science                   //
//                   University of Cape Town                      //
//                       Rondebosch 7700                          //
//                        SOUTH AFRICA                            //
////////////////////////////////////////////////////////////////////
// email: william@cs.uct.ac.za             phone: +27 21 650 2670 //
////////////////////////////////////////////////////////////////////

//#define VERBOSE
//#define DEBUG

using namespace std;

#include <iostream>
#include <iomanip>
#include <fstream>
#include <cstdio>
#include <cstdlib>
#include <cstdarg>
#include <cstring>
#include <cassert>
#include <cctype>
#include "common.hpp"
#include "dyna.hpp"
#include "info.hpp"
#include "token.hpp"
#include "parse.hpp"
#include "filenames.hpp"

#define UNDEFINED -1 

static char *simpleTypeTable[] = {
  "int", "char" , "long", "short", "long int", "short int", "unsigned", "unsigned char", "unsigned long", "unsigned short", NULL
};

Parser::Parser(char* basename) {
  tokeniser = new Tokeniser;
  info = new Information;
  advancedType = 0;
  elements = 0;
  transitions = 0;
  identifiers = 0;
  constants = 0;
  invariants = 0;
  stateMeasures = 0;
  countMeasures = 0;
  helpValues = 0;
  maxStates = 0;
  maxCpuTime = 0; 
  genRepStyle = solRepStyle = BRIEF;
  genRepInterval = 100000;
  solRepInterval = 5;
  method = AUTOMATIC;
  safeHash = 0;
  accuracy = 1e-10;
  maxIterations = 5000;
  dynamic = 1;
  relaxParameter = 1.0;
  doSteady = 0;
  doPassage = 0;
  doTransient = 0;

  
  char* stateListFileName   = add_extension(basename, "STATE") ;
  char* steadyStateFileName = add_extension(basename, "STEADY") ;
  char* matrixFileName      = add_extension(basename, "MATRIX") ;
  char* resultsFileName     = add_extension(basename, "RESULTS") ;

  strcpy(stateListFile, stateListFileName);
  strcpy(steadyStateFile, steadyStateFileName);
  strcpy(transitionMatrixFile, matrixFileName);
  strcpy(resultsFile, resultsFileName);
  strcpy(startVectorFile, "");
}

Parser::~Parser() {
  destroy();
}

void Parser::destroy() {
  int n;
  delete tokeniser;
  delete info;
  for (n=0; n<identifiers; n++) delete store[n];
  for (n=0; n<constants; n++) delete cons[n];
  for (n=0; n<invariants; n++) delete invar[n];
  for (n=0; n<transitions; n++) delete trans[n];
  for (n=0; n<stateMeasures; n++) delete statem[n];
  for (n=0; n<countMeasures; n++) delete countm[n];
  for (n=0; n<elements; n++) {
    delete type[n];
    delete instance[n];
  }
}

void Parser::terminate(char *fmt, ...)
{
  char buffer[2048];

  va_list argp;
  va_start(argp, fmt);
  vsprintf(buffer, fmt, argp); 
  va_end(argp);
  destroy();
  printf("\nfatal error: %s.\nexecution halted.\n", buffer);  
  exit(1);
}

int Parser::simpleType(char *_type) {
  for (long n=0; simpleTypeTable[n]; n++)
    if (!strcmp(_type, simpleTypeTable[n])) return 1;
  return 0;
}

void Parser::addStateVectorElement(char *_type, char *_instance) {
  if (!simpleType(_type)) advancedType = 1;
  type.check(elements);
  instance.check(elements);
  type[elements] = new char [strlen(_type) + 1];
  instance[elements] = new char [strlen(_instance) + 1];
  strcpy(type[elements], _type);
  strcpy(instance[elements], _instance);
  insert(_instance);
  elements++;
}

void Parser::mark(char *text) {
#ifdef LOGGING
  cout << tokeniser->getLineNumber() << ": " << text << endl;
#endif
}

int Parser::model() {
  Token token;
  mark("model");
  if (!tokeniser->expect(T_OPEN_BRACE)) return 0;
  token = tokeniser->getNextToken();
  while (token != T_CLOSE_BRACE && token != T_EOF) {
    switch (token) {
    case T_STATE_VECTOR:
      stateVector();
      break;
    case T_HEADER:
      tokeniser->scanCodeBlock(includeCB);
      break;
    case T_INITIAL:
      tokeniser->scanCodeBlock(initialCB, 2);
      break;
    case T_TRANSITION:
      transition();
      break;
    case T_CONSTANT:
      constant();
      break;  
    case T_INVARIANT:
      invariant();
      break;
    case T_HELP_VALUE:
      helpValue();
      break;
    case T_PRIMARY:
      mark("primary hash value function");
      tokeniser->scanCodeBlock(primaryCB, 2);
      break;
    case T_SECONDARY:
      mark("secondary hash value function");
      tokeniser->scanCodeBlock(secondaryCB, 2);
      break;
    case T_OUTPUT:
      mark("local output function");
      tokeniser->scanCodeBlock(outputCB, 2);
      break;
    default:
      tokeniser->unexpected(token, "model description");
      break;
    }
    token = tokeniser->getNextToken();
  }
  return 1;
}

int Parser::passage() {
  Token token;
  if(doPassage)
    mark("passage time");
  else if(doTransient)
    mark("transient measure");

  if (!tokeniser->expect(T_OPEN_BRACE)) return 0;
  token = tokeniser->getNextToken();
  while (token != T_CLOSE_BRACE && token != T_EOF) {
    switch(token) {
    case T_SOURCE_CONDITION:
      mark("initial state weighting function");
      tokeniser->scanCodeBlock(weightCB, 2);
      break;
    case T_TARGET_CONDITION:
      mark("target state condition");
      tokeniser->scanCodeBlock(targetCB);
      break;
    case T_EXCLUDE_CONDITION:
      mark("excluded state condition");
      tokeniser->scanCodeBlock(excludedCB);
      break;
    case T_START:
      mark("t_start");
      if (!tokeniser->expect(T_OPEN_BRACE)) return 0;
      if (tokeniser->expect(T_IDENTIFIER)) {
        t_start = atof(tokeniser->getIdentifier());
      }
      if (!tokeniser->expect(T_CLOSE_BRACE)) tokeniser->find(T_CLOSE_BRACE);

      break;
    case T_STOP:
      mark("t_stop");
      if (!tokeniser->expect(T_OPEN_BRACE)) return 0;
      if (tokeniser->expect(T_IDENTIFIER)) {
        t_stop = atof(tokeniser->getIdentifier());
      }
      if (!tokeniser->expect(T_CLOSE_BRACE)) tokeniser->find(T_CLOSE_BRACE);
      break;
    case T_STEP:
      mark("t_step");
      if (!tokeniser->expect(T_OPEN_BRACE)) return 0;
      if (tokeniser->expect(T_IDENTIFIER)) {
        t_step = atof(tokeniser->getIdentifier());
      }
      if (!tokeniser->expect(T_CLOSE_BRACE)) tokeniser->find(T_CLOSE_BRACE);
      break;
    default:
      tokeniser->unexpected(token, "passage time specification");
      break;
    }
    token = tokeniser->getNextToken();
  }

  if(t_start==0 || t_stop==0 || t_step==0)
    terminate("t-range not specified");

  return 1;
}

int Parser::instanceList(char *currentType)
{
  if (!tokeniser->expect(T_OPEN_BRACE)) return 0;
  Token token;
  int more = 1;
  while (more) {
    if (!tokeniser->expect(T_IDENTIFIER)) return tokeniser->find(T_CLOSE_BRACE);
    addStateVectorElement(currentType, tokeniser->getIdentifier());
    token = tokeniser->getNextToken();
    switch(token) {
    case T_COMMA:
      break;
    case T_CLOSE_BRACE:
      more = 0;
      break;
    default:
      tokeniser->unexpected(token, "state vector");
      break;
    }
  }
  return 1;
}

int Parser::stateVector()
{
  CodeBlock currentType;
  mark("parsing state vector");
  Token token;
  if (!tokeniser->expect(T_OPEN_BRACE)) return 0;
  token = tokeniser->getNextToken();
  while (token != T_CLOSE_BRACE && token != T_EOF) {
    switch(token) {
    case T_TYPE:
      tokeniser->scanCodeBlock(currentType);
      if (currentType.getArray()) 
	instanceList(currentType.getArray());
      break;
    default:
      tokeniser->unexpected(token, "state vector");
    }
    token = tokeniser->getNextToken();
  }
  return 1;
}

int Parser::insert(char *id) {
  for (int n=0; n < identifiers; n++)
    if (!strcmp(id, store(n)))
      return tokeniser->error("Duplicate identifier '%s'", id);
  store.check(identifiers);
  store[identifiers] = new char[strlen(id) + 1];
  strcpy(store[identifiers], id);
  identifiers++;
  return 1;
}

int Parser::transition() {
  Token token;

  mark("parsing transition");
  if (!tokeniser->expect(T_OPEN_BRACE)) return 0;
  if (!tokeniser->expect(T_IDENTIFIER)) return tokeniser->find(T_CLOSE_BRACE);
  char *identity = tokeniser->getIdentifier();
//  if (!insert(identity)) return 0;
  insert(identity);

  trans.check(transitions);
  Transition *current = new Transition;
  trans[transitions] = current;
  transitions++;

  current->setIdentity(identity);
  current->setType(TIMED);
  current->setPriority(0);

  if (!tokeniser->expect(T_CLOSE_BRACE)) tokeniser->find(T_CLOSE_BRACE);

  if (!tokeniser->expect(T_OPEN_BRACE)) return 0;
  token = tokeniser->getNextToken();
  while (token != T_CLOSE_BRACE && token != T_EOF) {
    switch(token) {
    case T_CONDITION:
      mark("condition"); 
      tokeniser->scanCodeBlock(current->getCondition());
      break; 
    case T_ACTION:
      mark("action");
      tokeniser->scanCodeBlock(current->getAction(), 4);
      break;
    case T_RATE:
      mark("rate");
      current->setType(TIMED);
      tokeniser->scanCodeBlock(current->getWeight());
      break;
    case T_WEIGHT:
      mark("weight");
      current->setType(INSTANTANEOUS);
      tokeniser->scanCodeBlock(current->getWeight());
      break;
    case T_PRIORITY:
    {
      mark("priority");
      if (!tokeniser->expect(T_OPEN_BRACE)) return 0;
      tokeniser->scanIdentifier();
      int prior = atoi(tokeniser->getIdentifier());
      if (tokeniser->checkInteger() && prior >= 0)
        current->setPriority(prior);
      else
	tokeniser->error("Invalid priority value '%s' (must be positive integer)", tokeniser->getIdentifier());
      if (!tokeniser->expect(T_CLOSE_BRACE)) tokeniser->find(T_CLOSE_BRACE);
      break;
    }
    case T_SOJOURNTIME_LT:
      mark("sojourn time laplace transform");
      tokeniser->scanCodeBlock(current->getSojourn());
      break;
    case T_SOJOURNTIME:
      mark("sojourn time distribution");
      tokeniser->scanCodeBlock(current->getDistribution());
      break;
    default:
      tokeniser->unexpected(token, "transition description");
      break;
    }
    token = tokeniser->getNextToken();
  }
  return 1;
}

int Parser::constant() {
  mark("constant");
  if (!tokeniser->expect(T_OPEN_BRACE)) return 0;
  if (!tokeniser->expect(T_IDENTIFIER)) return tokeniser->find(T_CLOSE_BRACE);

  char *identity = tokeniser->getIdentifier();
//  if (!insert(identity)) return 0;
  insert(identity);

  cons.check(constants);
  Constant *current = new Constant;
  cons[constants] = current;
  constants++;

  current->setIdentity(identity);

  if (!tokeniser->expect(T_CLOSE_BRACE)) tokeniser->find(T_CLOSE_BRACE);

  return tokeniser->scanCodeBlock(current->getDefine());
}

int Parser::helpValue() {
  CodeBlock currentType;

  mark("help value");
//  if (!tokeniser->expect(T_OPEN_BRACE)) return 0;
  tokeniser->scanCodeBlock(currentType);
  if (!currentType.getArray()) return 0;
//  if (!tokeniser->expect(T_IDENTIFIER)) return tokeniser->find(T_CLOSE_BRACE);
  
  helpv.check(helpValues);
  HelpValue *current = new HelpValue;
  helpv[helpValues] = current;
  helpValues++;

  current->setType(currentType.getArray());
//  if (!tokeniser->expect(T_CLOSE_BRACE)) tokeniser->find(T_CLOSE_BRACE);

  if (!tokeniser->expect(T_OPEN_BRACE)) return 0;
  if (!tokeniser->expect(T_IDENTIFIER)) return tokeniser->find(T_CLOSE_BRACE);

  char *identity = tokeniser->getIdentifier();
  insert(identity);
  current->setIdentity(identity);
  if (!tokeniser->expect(T_CLOSE_BRACE)) tokeniser->find(T_CLOSE_BRACE);

  return tokeniser->scanCodeBlock(current->getExpression());
}

int Parser::invariant() {
  mark("invariant");
/*  if (!tokeniser->expect(T_OPEN_BRACE)) return 0;
  if (!tokeniser->expect(T_IDENTIFIER)) return tokeniser->find(T_CLOSE_BRACE);

  char *identity = tokeniser->getIdentifier();
  insert(identity);
*/
  invar.check(invariants);
  Invariant *current = new Invariant;
  invar[invariants] = current;
  invariants++;

//  current->setIdentity(identity);

//  if (!tokeniser->expect(T_CLOSE_BRACE)) tokeniser->find(T_CLOSE_BRACE);

  return tokeniser->scanCodeBlock(current->getExpression());
}

int Parser::generation() {
  Token token;

  mark("generation");
  if (!tokeniser->expect(T_OPEN_BRACE)) return 0;
  token = tokeniser->getNextToken();
  while (token != T_CLOSE_BRACE && token != T_EOF) {
    switch(token) {
    case T_MAX_STATES:
      mark("max states");
      if (!tokeniser->expect(T_OPEN_BRACE)) return 0;
      if (tokeniser->expect(T_IDENTIFIER)) {
        maxStates = atol(tokeniser->getIdentifier());
        if (maxStates <= 0)
          tokeniser->error("Maximum states must be greater than 0 (was %d)", maxStates);
      }
      if (!tokeniser->expect(T_CLOSE_BRACE)) tokeniser->find(T_CLOSE_BRACE);
      break; 
    case T_SAFE_HASH:
      mark("safe hash");
      if (!tokeniser->expect(T_OPEN_BRACE)) return 0;
      token = tokeniser->getNextToken();
      if ( (token != T_ON) && (token != T_OFF) )
        tokeniser->unexpected(token, "generation control");
      else
	if (token == T_ON) safeHash = 1;
      if (!tokeniser->expect(T_CLOSE_BRACE)) tokeniser->find(T_CLOSE_BRACE);
      break;
    case T_MAX_CPU_TIME:
      mark("max cpu time");
      if (!tokeniser->expect(T_OPEN_BRACE)) return 0;
      if (tokeniser->scanReal() == T_IDENTIFIER) {
        maxCpuTime = atof(tokeniser->getIdentifier());
        if (maxCpuTime <= 0)
          tokeniser->error("Maximum CPU time must be greater than 0 (was %lf)", maxCpuTime);
      }
      if (!tokeniser->expect(T_CLOSE_BRACE)) tokeniser->find(T_CLOSE_BRACE);
      break;
    case T_REPORT_STYLE:
      mark("report style");
      if (!tokeniser->expect(T_OPEN_BRACE)) return 0;
      token = tokeniser->getNextToken();
      switch(token) {
      case T_FULL  : genRepStyle = FULL; break;
      case T_BRIEF : genRepStyle = BRIEF; break;
      case T_NONE  : genRepStyle = NONE; break;
      default:
        tokeniser->unexpected(token, "report style");
      }
      if (!tokeniser->expect(T_CLOSE_BRACE)) tokeniser->find(T_CLOSE_BRACE);
      break;
    case T_REPORT_INTERVAL:
      mark("report interval");
      if (!tokeniser->expect(T_OPEN_BRACE)) return 0;
      if (!tokeniser->expect(T_IDENTIFIER)) return tokeniser->find(T_CLOSE_BRACE);
      genRepInterval = atol(tokeniser->getIdentifier());
      if (genRepInterval <= 0)
        tokeniser->error("Reporting interval must be greater than 0 (was %ld)", genRepInterval);
      if (!tokeniser->expect(T_CLOSE_BRACE)) tokeniser->find(T_CLOSE_BRACE);
      break;
    default:
      tokeniser->unexpected(token, "generation control");
      break;
    }
    token = tokeniser->getNextToken();
  }
  return 1;
}

int Parser::solution() {
  Token token;

  mark("solution");
  if (!tokeniser->expect(T_OPEN_BRACE)) return 0;
  token = tokeniser->getNextToken();
  while (token != T_CLOSE_BRACE && token != T_EOF) {
    switch(token) {
    case T_METHOD:
      if (!tokeniser->expect(T_OPEN_BRACE)) return 0;
      token = tokeniser->getNextToken();
      switch(token) {
      case T_GAUSS        : method = GAUSS; break;
      case T_LU           : method = LU; break;
      case T_GRASSMANN    : method = GRASSMANN; break;
      case T_POWER        : method = POWER; break;
      case T_GAUSS_SEIDEL : method = GAUSS_SEIDEL; break;
      case T_SOR          : method = SOR; break;
      case T_CGNR         : method = CGNR; break;
      case T_CGS          : method = CGS; break;
      case T_BICG         : method = BICG; break;
      case T_BICGSTAB     : method = BICGSTAB; break;
      case T_BICGSTAB2    : method = BICGSTAB2; break;
      case T_AI           : method = AI; break;
      case T_AIR          : method = AIR; break;
      case T_MYCG         : method = MYCG; break;
      case T_AUTOMATIC    : method = AUTOMATIC; break;
      default:
        tokeniser->unexpected(token, "solution method");
      }
      if (!tokeniser->expect(T_CLOSE_BRACE)) tokeniser->find(T_CLOSE_BRACE);
      break;
    case T_ACCURACY:
      mark("accuracy");
      if (!tokeniser->expect(T_OPEN_BRACE)) return 0;
      if (tokeniser->scanReal() == T_IDENTIFIER) {
        accuracy = atof(tokeniser->getIdentifier());
        if (accuracy <= EPS || accuracy > 0.01)
	  tokeniser->error("Accuracy must be between %s and %s (was %lf)", EPS_STR, "0.01", accuracy);	
      }
      if (!tokeniser->expect(T_CLOSE_BRACE)) tokeniser->find(T_CLOSE_BRACE);
      break;
    case T_MAX_ITERATIONS:
      mark("max iterations");
      if (!tokeniser->expect(T_OPEN_BRACE)) return 0;
      if (tokeniser->expect(T_IDENTIFIER)) {
        maxIterations = atol(tokeniser->getIdentifier());
        if (maxIterations < 0)
          tokeniser->error("Maximum iterations must be >= 0 (was %ld)", maxIterations);
      }
      if (!tokeniser->expect(T_CLOSE_BRACE)) tokeniser->find(T_CLOSE_BRACE);
      break;
    case T_RELAX_PARAMETER:
      mark("relax parameter");
      if (!tokeniser->expect(T_OPEN_BRACE)) return 0;
      token = tokeniser->scanReal();
      if (token == T_DYNAMIC) {
        dynamic = 1;
      } else if (token == T_IDENTIFIER) {
        dynamic = 0;
        relaxParameter = atof(tokeniser->getIdentifier());
        if (relaxParameter < 0 || relaxParameter > 2)
          tokeniser->error("Relaxation parameter should lie in interval from 0 to 2 (was %lf)", relaxParameter);
      }
      if (!tokeniser->expect(T_CLOSE_BRACE)) tokeniser->find(T_CLOSE_BRACE);
      break;
    case T_START_VECTOR:
      mark("start vector");
      tokeniser->scanCodeBlock(startVectorCB);
      if (startVectorCB.getArray())
	strcpy(startVectorFile, startVectorCB.getArray());
      break; 
    case T_REPORT_STYLE:
      mark("solution style");
      if (!tokeniser->expect(T_OPEN_BRACE)) return 0;
      token = tokeniser->getNextToken();
      switch(token) {
      case T_FULL  : solRepStyle = FULL; break;
      case T_BRIEF : solRepStyle = BRIEF; break;
      case T_NONE  : solRepStyle = NONE; break;
      default:
        tokeniser->unexpected(token, "report style");
        break;
      }
      if (!tokeniser->expect(T_CLOSE_BRACE)) tokeniser->find(T_CLOSE_BRACE);
      break;
    case T_REPORT_INTERVAL:
      mark("solution interval");
      if (!tokeniser->expect(T_OPEN_BRACE)) return 0;
      if (!tokeniser->expect(T_IDENTIFIER)) return tokeniser->find(T_CLOSE_BRACE);
      solRepInterval = atol(tokeniser->getIdentifier());
      if (solRepInterval <= 0)
        tokeniser->error("Reporting interval must be greater than 0 (was %ld)", solRepInterval);
      if (!tokeniser->expect(T_CLOSE_BRACE)) tokeniser->find(T_CLOSE_BRACE);
      break;
    default:
      tokeniser->unexpected(token, "solution control");
      break;
    }
    token = tokeniser->getNextToken();
  }
  return 1;
}

int Parser::stateMeasure()
{
  CodeBlock identity;
  Token token, estToken;
  mark("state measure");
  tokeniser->scanCodeBlock(identity);
//  if (!tokeniser->expect(T_OPEN_BRACE)) return 0;
//  if (!tokeniser->expect(T_IDENTIFIER)) return tokeniser->find(T_CLOSE_BRACE);

//  char *identity = tokeniser->getIdentifier();
//  if (!insert(identity)) return 0;
  insert(identity.getArray());

  statem.check(stateMeasures);
  StateMeasure *current = new StateMeasure;
  statem[stateMeasures] = current;
  stateMeasures++;

  current->setIdentity(identity.getArray());

//  if (!tokeniser->expect(T_CLOSE_BRACE)) tokeniser->find(T_CLOSE_BRACE);

  if (!tokeniser->expect(T_OPEN_BRACE)) return 0;
  token = tokeniser->getNextToken();
  while (token != T_CLOSE_BRACE && token != T_EOF) {
    switch(token) {
    case T_ESTIMATOR:
      mark("estimator"); 
      if (!tokeniser->expect(T_OPEN_BRACE)) return 0;
      estToken = tokeniser->getNextToken();
      while (estToken != T_CLOSE_BRACE) {
        switch(estToken) {
        case T_MEAN         : current->addEstimator(MEAN); break;
        case T_VARIANCE     : current->addEstimator(VARIANCE); break;
        case T_STDDEV       : current->addEstimator(STDDEV); break;
        case T_DISTRIBUTION : current->addEstimator(DISTRIBUTION); break;
        case T_COMMA        : break;
	default:
          tokeniser->unexpected(estToken, "estimator");
          break;
	}
        estToken = tokeniser->getNextToken();
      }
      break; 
    case T_EXPRESSION:
      mark("expression");
      tokeniser->scanCodeBlock(current->getExpression());
      break;
    default:
      tokeniser->unexpected(token, "state measure");
      break;
    }
    token = tokeniser->getNextToken();
  }
  return 1;
}

int Parser::findTransitionIndex(char *target)
{
  int n;

  for (n=0; n<transitions; n++)
    if (!strcmp (target,trans[n]->getIdentity())) return n;
  return UNDEFINED;
}

int Parser::transitionList(CountMeasure *current) 
{
  Token token;
  int index;
  char *id;

  mark("transition list");
  if (!tokeniser->expect(T_OPEN_BRACE)) return 0;
  token = tokeniser->getNextToken();
  while (token != T_CLOSE_BRACE && token != T_EOF) {
    switch (token) {
    case T_ALL:
      current->setAll(1);
      break;
    case T_IDENTIFIER:
      id = tokeniser->getIdentifier();
      index = findTransitionIndex(id);
      if (index==UNDEFINED) {
        tokeniser->error("Unknown transition '%s' in transition list for count measure '%s'", id, current->getIdentity());
      } else {
        if (trans(index)->getType() == INSTANTANEOUS)
          tokeniser->warning("instantaneous transition '%s' in transition list for count measure '%s' will be ignored by performance analyser", id, current->getIdentity());
        current->addTransition(index);
      }
      break;
    case T_COMMA:
      break;
    default:
      tokeniser->unexpected(token, "transition list");
      break;
    }
    token = tokeniser->getNextToken();
  }
  return 1;
}

int Parser::countMeasure()
{
  CodeBlock identity;
  Token token, estToken;
  mark("count measure");
  tokeniser->scanCodeBlock(identity);
/*  if (!tokeniser->expect(T_OPEN_BRACE)) return 0;
  if (!tokeniser->expect(T_IDENTIFIER)) return tokeniser->find(T_CLOSE_BRACE);

  char *identity = tokeniser->getIdentifier();*/
//  if (!insert(identity)) return 0;
  insert(identity.getArray());

  countm.check(countMeasures);
  CountMeasure *current = new CountMeasure;
  countm[countMeasures] = current;
  countMeasures++;

  current->setIdentity(identity.getArray());

//  if (!tokeniser->expect(T_CLOSE_BRACE)) tokeniser->find(T_CLOSE_BRACE);

  if (!tokeniser->expect(T_OPEN_BRACE)) return 0;
  token = tokeniser->getNextToken();
  while (token != T_CLOSE_BRACE && token != T_EOF) {
    switch(token) {
    case T_ESTIMATOR:
      mark("estimator"); 
      if (!tokeniser->expect(T_OPEN_BRACE)) return 0;
      estToken = tokeniser->getNextToken();
      while (estToken != T_CLOSE_BRACE) {
        switch(estToken) {
        case T_MEAN         : current->addEstimator(MEAN); break;
        case T_COMMA        : break;
        case T_VARIANCE     : tokeniser->error("Variance of count measures unavailable"); break;
	default:
          tokeniser->unexpected(estToken, "estimator");
          break;
	}
        estToken = tokeniser->getNextToken();
      }
      break; 
    case T_PRECONDITION:
      mark("precondition");
      tokeniser->scanCodeBlock(current->getPrecondition());
      break;
    case T_POSTCONDITION:
      mark("postcondition");
      tokeniser->scanCodeBlock(current->getPostcondition());
      break;
    case T_TRANSITION:
      transitionList(current);
      break;
    default:
      tokeniser->unexpected(token, "count measure");
      break;
    }
    token = tokeniser->getNextToken();
  }
  return 1;
}

int Parser::performance() {
  Token token;
  mark("parsing performance");
  if (!tokeniser->expect(T_OPEN_BRACE)) return 0;
  token = tokeniser->getNextToken();
  while (token != T_CLOSE_BRACE && token != T_EOF) {
    switch (token) {
    case T_STATE_MEASURE:
      stateMeasure();
      break;
    case T_COUNT_MEASURE:
      countMeasure();
      break;
    default:
      tokeniser->unexpected(token, "performance measures");
      break;
    }
    token = tokeniser->getNextToken();
  }
  return 1;
}

int Parser::output() {
  Token token;

  mark("output options");
  if (!tokeniser->expect(T_OPEN_BRACE)) return 0;
  token = tokeniser->getNextToken();
  while (token != T_CLOSE_BRACE && token != T_EOF) {
    switch(token) {
    case T_STATELIST:
      mark("state list");
      if (!tokeniser->expect(T_OPEN_BRACE)) return 0;
      if (tokeniser->expect(T_IDENTIFIER))
        strcpy(stateListFile, tokeniser->getIdentifier());
      if (!tokeniser->expect(T_CLOSE_BRACE)) tokeniser->find(T_CLOSE_BRACE);
      break; 
    case T_STEADY_STATE_VECTOR:
      mark("steady state vector");
      if (!tokeniser->expect(T_OPEN_BRACE)) return 0;
      if (tokeniser->expect(T_IDENTIFIER))
        strcpy(steadyStateFile, tokeniser->getIdentifier());
      if (!tokeniser->expect(T_CLOSE_BRACE)) tokeniser->find(T_CLOSE_BRACE);
      break;
    case T_TRANSITION_MATRIX:
      mark("transition matrix");
      if (!tokeniser->expect(T_OPEN_BRACE)) return 0;
      if (tokeniser->expect(T_IDENTIFIER))
        strcpy(transitionMatrixFile, tokeniser->getIdentifier());
      if (!tokeniser->expect(T_CLOSE_BRACE)) tokeniser->find(T_CLOSE_BRACE);
      break;
    case T_PERFORMANCE_RESULTS:
      mark("performance results");
      if (!tokeniser->expect(T_OPEN_BRACE)) return 0;
      if (tokeniser->expect(T_IDENTIFIER)) 
        strcpy(resultsFile, tokeniser->getIdentifier());
      if (!tokeniser->expect(T_CLOSE_BRACE)) tokeniser->find(T_CLOSE_BRACE);
      break;
    default:
      tokeniser->unexpected(token, "output options");
      if (!tokeniser->expect(T_CLOSE_BRACE)) tokeniser->find(T_CLOSE_BRACE);
      break;
    }
    token = tokeniser->getNextToken();
  }
  return 1;
}

int Parser::parse(char *filename, char* basename) {
  Token token;

  if (!tokeniser->openSource(filename))
    terminate("parse.cxx cannot open file '%s'", filename);

  strcpy(sourceFile, filename);
  
  token = tokeniser->getNextToken();
  while (token != T_EOF) {
    switch(token) {
    case T_MODEL_DESCRIPTION:
      model();
      break;
    case T_GENERATION_CONTROL:
      generation();
      break;
    case T_SOLUTION_CONTROL:
      solution();
      break;
    case T_PERFORMANCE:
      if(doTransient || doPassage)
	terminate("both steady-state and passage/transient measures specified, please choose one");
      doSteady = 1;
      doPassage = 0;
      doTransient = 0;
      performance();
      break;
    case T_PASSAGE:
      if(doTransient || doSteady)
	terminate("multiple types of performance measures specified, please choose one");
      doPassage = 1;
      doSteady = 0;
      doTransient = 0;
      passage();
      break;
    case T_TRANSIENT:
      if(doPassage || doSteady)
	terminate("multiple types of performance measures specified, please choose one");
      doTransient = 1;
      doSteady = 0;
      doPassage = 0;
      passage();
      break;
    case T_OUTPUT:
      output();
      break;
    default:
      tokeniser->unexpected(token, "top level parsing");
      break;
    }
    token = tokeniser->getNextToken();
  }
  if (tokeniser->getErrors()) {
    tokeniser->report();
    return 0;
  }

  char* userCxxname = add_extension(basename, "user.cxx") ;
  char* optionsname = add_extension(basename, "OPTIONS" ) ;

  /* 
  string basename( remove_extension(filename) ) ;
  string userCxxname = basename + "." + "user.cxx" ;
  string optionsname = basename + "." + "OPTIONS"  ;
   */

  generate( userCxxname ) ;
  tokeniser->report();

  if (tokeniser->getErrors()) return 0;

  setInformation();
  if (!info->output( optionsname ) )
    terminate("cannot update options file");
//  info->input("OPTIONS");
  return 1;
}

void Parser::setInformation() {
  int n;
  info->setTransitions(transitions);
  for (n=0; n<transitions; n++) {
    info->setTransType(n, trans(n)->getType());
    info->setPriority(n, trans(n)->getPriority());
  }
  info->setStateMeasures(stateMeasures);
  info->setCountMeasures(countMeasures);
  info->setMaxStates(maxStates);
  info->setMaxIterations(maxIterations);
  info->setGenRepInterval(genRepInterval);
  info->setSolRepInterval(solRepInterval);
  info->setMaxCPUTime(maxCpuTime);
  info->setSafeHash(safeHash);
  info->setAccuracy(accuracy);
  info->setRelaxParameter(relaxParameter);
  info->setDynamic(dynamic);
  info->setGenRepStyle(genRepStyle);
  info->setSolRepStyle(solRepStyle);
  info->setMethod(method);
  info->setStateListFile(stateListFile);
  info->setStartVectorFile(startVectorFile);
  info->setSteadyStateFile(steadyStateFile);
  info->setTransitionMatrixFile(transitionMatrixFile);
  int *stateEst = new int[stateMeasures];
  int *countEst = new int[countMeasures];
  
  for (n=0; n<stateMeasures; n++) {
    stateEst[n] = statem(n)->getEstimator();
    info->setStateName(n, statem(n)->getIdentity());
  }
  for (n=0; n<countMeasures; n++) {
    countEst[n] = countm(n)->getEstimator();
    info->setCountName(n, countm(n)->getIdentity());
  }
  info->setStateEstimator(stateEst);
  info->setCountEstimator(countEst);
  info->setTStart(t_start);
  info->setTStop(t_stop);
  info->setTStep(t_step);

  info->setDoSteady(doSteady);
  info->setPassage(doPassage);
  info->setTransient(doTransient);

  info->setParse(1);

  char command[256], mdcheck[128];
  sprintf(command,"md5sum %s", sourceFile);
  FILE *fp = popen(command, "r");
  fscanf(fp, "%s", mdcheck);
#ifdef LOGGING
  cout << "---> MD5 CHECKSUM IS " << mdcheck << "<--" << endl;
#endif
  info->setChecksum(mdcheck);

}

int main(int argc, char *argv[]) {

  if (argc < 3) {
    cout << argv[0] << ": input file *and* basename required." << endl;
    return 1;
  }
  
  Parser *honey = new Parser (argv[2]);
  int success = honey->parse(argv[1], argv[2]);
  delete honey;
  
  return !success;
}

