////////////////////////////////////////////////////////////////////
// WMCA Markov Chain Analyser        (C) 1995 William Knottenbelt //
////////////////////////////////////////////////////////////////////
//            Data Network Architectures Laboratory               //
//               Department of Computer Science                   //
//                   University of Cape Town                      //
//                       Rondebosch 7700                          //
//                        SOUTH AFRICA                            //
////////////////////////////////////////////////////////////////////
// email: william@cs.uct.ac.za             phone: +27 21 650 2670 //
////////////////////////////////////////////////////////////////////

// Update history: 7 May 1996  check SOR converging to neg. solution

using namespace std;

#include <iostream>
#include <iomanip>
#include <cassert>
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <complex>
#include "common.hpp"
#include "dyna.hpp"
#include "timer.hpp"
#include "info.hpp"
#include "element.hpp"
#include "vector.hpp"
#include "store.hpp"
#include "avl.hpp"
#include "sparse-nocomplex.hpp"
#include "hydra-steady.hpp"

// ### was 0.005
#define SOR_MOVE 0.005
#define SOR_MOVE_2 0.025

// and the winner is... table driven SOR...not!
// simple run up SOR


#ifndef LOGGING
#define SILENT 1
#endif

void SteadyStateAnalyser::power(Vector<double> &solution, long maxIterations, double accuracy) {
#ifndef SILENT 
  cout << "SteadyStateAnalyser (" << this << "): starting power solve..." << endl;
#endif
  Vector<double> last, result;
  last.setLength(tangible);
  result.setLength(tangible);
  int k,n;
  QC->print();
  solution.normalise();
  for (n=0; n<tangible; n++)
    last[n] = solution(n);

  double maxDiff, solNorm, convergence;
  int done = 0;
  //int changed;
  for (k=0; k<2000 && !done ; k++) {
    //    cout << "solution: "<< solution << endl;
    QC->preMultiply(solution, result);
    //    cout << "result: "<< result << endl;
    solNorm = 0;
    maxDiff = 0;
    for (n=0; n<tangible; n++) {
      solution[n] = result(n);
      if (fabs(solution(n)) > solNorm) solNorm = fabs(solution(n));
      if (fabs(solution(n) - last(n)) > maxDiff) maxDiff = fabs(solution(n) - last(n));
      last[n] = solution(n);    
    }
    convergence = maxDiff/solNorm;
    cout << "iteration = " << k << " " << convergence << endl;
    if (convergence < accuracy) done = 1;
  }
  cout << "EXPERIMENT " << k << endl;
  unScale(solution);
  solution.normalise();
  solution.output(steadyVectorFile);
}

void SteadyStateAnalyser::dynamicSOR(Vector<double> &solution, long maxIterations, double accuracy)
{
  #ifndef SILENT 
  cout << "SteadyStateAnalyser (" << this << "): starting dynamic SOR solve..." << endl;
  #endif

  maxIterations = 500000000;

  Vector<double> last;
  int k,m;
  double diffNorm, solNorm, solInfNorm, convergence = 1.0, omega = 1.0;

  // table-based SOR optimization method
  int count[81]; // index = 0, maxIndex = 0, stop = 0;
  double average[81], lastResidInfNorm, improve, step = 0.05;
  long n;

  for (n=0; n<81; n++) {
    count[n] = 0;
    average[n] = 0;
  }

#ifdef EXPERIMENT
  Vector<double> normSol, normRes;
  normSol.setLength(tangible);
  normRes.setLength(tangible);
#endif

  last.setLength(tangible);

  Q->calcResidual(solution, last);
  lastResidInfNorm = last.infNorm();

  for (n=0; n<tangible; n++)
    last[n] = solution(n);

#ifdef EXPERIMENT
  cout << "LAST RESID INF NORM = " << lastResidInfNorm << endl;
  for (n=0; n<tangible; n++)
    normSol[n] = solution[n];
  normSol.normalise();
  Q->calcResidual(normSol, normRes);
  cout << "iteration " << 0 << " infNorm=" << normRes.infNorm() << " twoNorm=" << normRes.twoNorm() << endl;
  cout << "EXPERIMENT " << 0 << " " << log10(normRes.twoNorm()) << endl;  
#endif

  timer->read();
#ifndef SILENT
  if (solRepStyle != NONE) {
    cout << "Iteration        Omega             Resid.-norm         Accuracy" << endl;
    cout << "----------------+-----------------+-------------------+-------------------" << endl;
  }
#endif
  cout.setf(ios::showpoint);
  long iterations = 0;
  do {
    m = 5; // steps(iterations);
    for (k=0; k<m; k++) {
      Q->SORIteration(solution, omega);
#ifdef EXPERIMENT
    for (n=0; n<tangible; n++)
      normSol[n] = solution[n];
    normSol.normalise();
    Q->transMultiply(normSol, normRes);
    cout << "iteration " << iterations+k+1 << " infNorm=" << normRes.infNorm() << " twoNorm=" << normRes.twoNorm() << endl;
    cout << "EXPERIMENT " << (iterations+k+1)*1.2 << " " << log10(normRes.twoNorm()) << endl;
#endif
    }
    Q->calcResidualNorm(solution, solNorm, solInfNorm, last, diffNorm);
    iterations += m;
    if (diffNorm != 0) convergence = diffNorm / solNorm;
      else convergence = 0;

    improve = (lastResidInfNorm-solInfNorm)/lastResidInfNorm;
    if (iterations == m) improve = 1.0;
    if (improve < 0) {
      omega = 1+(omega - 1)/2;
      step = step * 0.5;
      if (step < 0.005) step = 0.005;
    } else {
      omega += step;
    }

/*
    if (iterations == m) improve = 1.0;
// ###
    cout << "impr = " << improve << endl;
    if (!count[index])
      average[index] = improve;
    else 
      average[index] = 0.4*average[index] + 0.6*improve;
    if (average[index] > max) {
      max = average[index];
      maxIndex = index;
    } else {
      max = 0;
      for (n=0; n<81; n++)
	if (average[n] > max) {
	  max = average[n];
	  maxIndex = n;
	}
    }
// ###
    cout << "average[" << index << "] = " << average[index] << endl;
    count[index]++;
    if (improve < 0 && iterations > 5) stop = 1;
    if (iterations < 200 && !stop) {
      omega += 0.025;
      index += 2;
    }
    else {
      omega = (0.0125*maxIndex) + 1.0;
      index = maxIndex;
      if (index<80 && count[index+1]==0) {
	omega += 0.0125;
	index++;
      }
    }
    if (omega > 2.0) {
      omega = 2.0;
      index = 80;
    }
*/    
    for (n=0; n<tangible; n++)
      last[n] = solution(n);

#ifndef SILENT
    if (solRepStyle != NONE) {
      cout << setw(15) << iterations << " | " << setw(15) << omega << " | " << setw(17) << solInfNorm << " | " << setw(17) << convergence << endl;
    }
#endif
    lastResidInfNorm = solInfNorm;
  } while (iterations < maxIterations && convergence > accuracy);
  cout.unsetf(ios::showpoint);
  // check for convergence to negative solution
  checkNegative(solution);
  solution.normalise();
  Q->calcResidual(solution, last);
#ifndef SILENT
  if (solRepStyle != NONE) {
    cout << "----------------+-----------------+-------------------+-------------------" << endl;
    checkVector(solution, last.infNorm());
  }
#endif
  if (convergence > accuracy)
    terminate("dynamic SOR failed to converge within %d iterations", maxIterations);

  unScale(solution);
  solution.normalise();
#ifdef DEBUG
  cout << "result's infNorm: " << solInfNorm << endl;
#endif
  #ifndef RELEASE
  solution.textOutput("fixedsor.sol");
  #endif
  solution.output(steadyVectorFile);

  #ifndef SILENT 
  double timeUsed = timer->read();
  cout << "SteadyStateAnalyser (" << this << "): dynamic SOR complete (" << timeUsed << " CPU seconds)." << endl;
  #endif
}

void SteadyStateAnalyser::sor(double omega, Vector<double> &solution, long maxIterations, double accuracy)
{
  #ifndef SILENT 
  cout << "SteadyStateAnalyser (" << this << "): starting fixed SOR solve..." << endl;
  #endif

  Vector<double> last;
  int k,m;
  double diffNorm, solNorm, solInfNorm, convergence;
#ifdef EXPERIMENT
  Vector<double> normSol, normRes;
  normSol.setLength(tangible);
  normRes.setLength(tangible);
#endif

  last.setLength(tangible);

  long n;
  for (n=0; n<tangible; n++)
    last[n] = solution(n);

#ifdef EXPERIMENT
    for (n=0; n<tangible; n++)
      normSol[n] = solution[n];
    normSol.normalise();
    Q->calcResidual(normSol, normRes);
    cout << "iteration " << 0 << " infNorm=" << normRes.infNorm() << " twoNorm=" << normRes.twoNorm() << endl;
    cout << "EXPERIMENT " << 0 << " " << log10(normRes.twoNorm()) << endl;  
#endif

  timer->read();
  if (solRepStyle != NONE) {
    cout << "Iteration        Omega             Resid.-norm         Accuracy" << endl;
    cout << "----------------+-----------------+-------------------+-------------------" << endl;
  }
  cout.setf(ios::showpoint);
  long iterations = 0;
  do {
    m = 5; // steps(iterations);
    for (k=0; k<m; k++) {
      Q->SORIteration(solution, omega);
#ifdef EXPERIMENT
    for (n=0; n<tangible; n++)
      normSol[n] = solution[n];
    normSol.normalise();
    Q->transMultiply(normSol, normRes);
    cout << "iteration " << iterations+k+1 << " infNorm=" << normRes.infNorm() << " twoNorm=" << normRes.twoNorm() << endl;
    cout << "EXPERIMENT " << (iterations+k+1)*1.2 << " " << log10(normRes.twoNorm()) << endl;
#endif
    }
    Q->calcResidualNorm(solution, solNorm, solInfNorm, last, diffNorm);
    iterations += m;
    if (diffNorm != 0) convergence = diffNorm / solNorm;
      else convergence = 0;
    for (n=0; n<tangible; n++)
      last(n) = solution(n);

    if (solRepStyle != NONE) {
      cout << setw(15) << iterations << " | " << setw(15) << omega << " | " << setw(17) << solInfNorm << " | " << setw(17) << convergence << endl;
    }
  } while (iterations < maxIterations && convergence > accuracy);
  cout.unsetf(ios::showpoint);
  checkNegative(solution);
  solution.normalise();
  Q->calcResidual(solution, last);
  if (solRepStyle != NONE) {
    cout << "----------------+-----------------+-------------------+-------------------" << endl;
    checkVector(solution, last.infNorm());
  }
  if (convergence > accuracy)
    terminate("fixed SOR failed to converge within %d iterations", maxIterations);

  unScale(solution);
  solution.normalise();
#ifdef DEBUG
  cout << "result's infNorm: " << solInfNorm << endl;
#endif
  #ifndef RELEASE
  solution.textOutput("fixedsor.sol");
  #endif
  solution.output(steadyVectorFile);

  #ifndef SILENT 
  double timeUsed = timer->read();
  cout << "SteadyStateAnalyser (" << this << "): fixed SOR complete (" << timeUsed << " CPU seconds)." << endl;
  #endif
}

void SteadyStateAnalyser::sor(Vector <double> &solution, long maxIterations, double accuracy)
{
  sor(relaxParameter, solution, maxIterations, accuracy);
}

void SteadyStateAnalyser::gaussSeidel(Vector <double> &solution, long maxIterations, double accuracy)
{
  sor(1.0, solution, maxIterations, accuracy);
}


int SteadyStateAnalyser::steps(long n)
{
  if (n >= 1000) return 50;
  if (n >= 500) return 20;
  if (n >= 100) return 10;
  return 5;
}
// the original dynamic SOR
/*
void SteadyStateAnalyser::dynamicSOR(Vector<double> &solution, long maxIterations, double accuracy)
{
  #ifndef SILENT 
  cout << "SteadyStateAnalyser (" << this << "): starting dynamic SOR solve..." << endl;
  #endif
  Vector<double> last, alternate, result;
  double omega = 1.0, lo = 1.0, hi = 2.0, omegaLo = lo, omegaHi = hi;
  double solNorm, solInfNorm, altNorm, altInfNorm, diffNorm, altDiffNorm;
  double convergence, lastConvergence = 1.0, altConvergence;

  alternate.setLength(tangible);
  last.setLength(tangible);
  result.setLength(tangible);
#ifdef EXPERIMENT
  Vector<double> normSol, normRes;
  normSol.setLength(tangible);
  normRes.setLength(tangible);
#endif

  long n;
  for (n=0; n<tangible; n++) {
    last[n] = solution(n);
  }
  int k;

  long iterations = 0;
#ifdef EXPERIMENT
    for (n=0; n<tangible; n++)
      normSol[n] = solution[n];
    normSol.normalise();
    Q->calcResidual(normSol, normRes);
    cout << "iteration " << 0 << " infNorm=" << normRes.infNorm() << " twoNorm=" << normRes.twoNorm() << endl;
    cout << "EXPERIMENT " << 0 << " " << log10(normRes.twoNorm()) << endl;  
#endif

  timer->read();
  if (solRepStyle != NONE) {
    cout << "Iteration  Omega       Lower       Upper       Resid.-norm   Accuracy" << endl;
    cout << "----------+-----------+-----------+-----------+-------------+-------------" << endl;
  }

  for (k=0; k<10; k++) {
    Q->SORIteration(solution, omega);
    iterations++;
#ifdef EXPERIMENT
    for (n=0; n<tangible; n++)
      normSol[n] = solution[n];
    normSol.normalise();
    Q->transMultiply(normSol, normRes);
    cout << "iteration " << iterations << " infNorm=" << normRes.infNorm() << " twoNorm=" << normRes.twoNorm() << endl;
    cout << "EXPERIMENT " << iterations*1.2 << " " << log10(normRes.twoNorm()) << endl;
#endif
  }

  Q->calcResidualNorm(solution, solNorm, solInfNorm, last, diffNorm);
  if (diffNorm != 0) convergence = diffNorm /solNorm;
    else convergence = 0;

  cout.setf(ios::showpoint);

  for (n=0; n<tangible; n++)
    last[n] = solution(n);

  int i,m;
  while (iterations < maxIterations && convergence > accuracy) {
    if (solRepStyle != NONE) {
      cout << setw(9) << iterations << " | " << setw(9) << omega << " | " << setw(9) << lo << " | " << setw(9) << hi << " | " << setw(11) << solInfNorm << " | " << setw(11) << convergence << endl;
    }
    omega = (lo+hi)/2;
    omegaLo = (lo+omega)/2;
    omegaHi = (omega+hi)/2;

    m = 5; //steps(iterations);
    for (n=0; n<tangible; n++)
      alternate[n] = solution(n);

    for (k=0; k<m; k++) {
      Q->twinSORIteration(solution, omegaLo, alternate, omegaHi);
#ifdef EXPERIMENT
    for (n=0; n<tangible; n++)
      normSol[n] = solution[n];
    normSol.normalise();
    Q->transMultiply(normSol, normRes);
    cout << "iteration " << iterations+k+1 << " infNorm=" << normRes.infNorm() << " twoNorm=" << normRes.twoNorm() << endl;
    cout << "EXPERIMENT " << (iterations+k+1)*1.2 << " " << log10(normRes.twoNorm()) << endl;
#endif
    }
    iterations += m;
    Q->twinCalcResidualNorm(solution, solNorm, solInfNorm, alternate, altNorm, altInfNorm, last, diffNorm, altDiffNorm);

    if (diffNorm != 0) convergence = diffNorm / solNorm;
      else convergence = 0;

    if (altDiffNorm != 0) altConvergence = altDiffNorm / altNorm;
      else altConvergence = 0;

#ifdef DEBUG
    cout << "twinCalc: solInfNorm is " << solInfNorm << " and altInfNorm is " << altInfNorm << endl;
    cout << "diffNorm is : " << diffNorm << ", solNorm is " << solNorm << endl;
    cout << "altDiffNorm is : " << altDiffNorm << ", altNorm is " << altNorm << endl;
    cout << "iteration " << iterations << endl;
    cout << "using " << omegaLo << " as omega yields " << convergence << endl;
    Q->transMultiply(solution, result);
    cout << "using " << omegaHi << " as omega yields " << altConvergence << endl;
    Q->transMultiply(alternate, result);
    cout << "alternate infnorm is " << result.infNorm() << endl;
    cout << "solution infnorm is " << result.infNorm() << endl;
#endif

    if (altInfNorm < solInfNorm) {
#ifdef DEBUG
      cout << omegaHi << " is the better choice!\n";
#endif
      lo = omegaLo - SOR_MOVE;
      hi += SOR_MOVE;
      double *array = solution.getArray();
      solution.setArray(alternate.getArray(),tangible);
      alternate.setArray(array,tangible);
      convergence = altConvergence;
    } else {
#ifdef DEBUG
      cout << omegaLo << " is the better choice!\n";
#endif
      hi = omegaHi + SOR_MOVE;
      lo -= SOR_MOVE;
    }

    if (solInfNorm > 1e06) {
      cout << "normalising..." << endl;
      lo = 1.0 - SOR_MOVE;
      hi = 1.0 + SOR_MOVE;
      for (n=0; n<tangible; n++)
        solution[n] = fabs(solution(n));
      solution.normalise(tangible);
    }

    if (hi > 2.0) hi == 2.0;

    for (n=0; n<tangible; n++) {
//      solution(n) = fabs(solution(n));
//      if (solution(n) < 0) solution(n) = 0;
      last[n] = solution(n);
    }

    lastConvergence = convergence;
  }

  solution.normalise();
  Q->calcResidual(solution, last);  
  if (solRepStyle != NONE) {
    cout << setw(9) << iterations << " | " << setw(9) << omega << " | " << setw(9) << lo << " | " << setw(9) << hi << " | " << setw(11) << solInfNorm << " | " << setw(11) << convergence << endl;
    cout << "----------+-----------+-----------+-----------+-------------+-------------" << endl;
    checkVector(solution, last.infNorm());
  }
  cout.unsetf(ios::showpoint);

  if (convergence > accuracy)
    terminate("dynamic SOR failed to converge within %d iterations", maxIterations);
  unScale(solution);
  solution.normalise();
  double timeUsed = timer->read();
#ifdef DEBUG
  cout << "result's infNorm: " << solInfNorm << endl;
#endif
  #ifndef RELEASE
  solution.textOutput("sor.sol");
  #endif
  solution.output(steadyVectorFile);

  #ifndef SILENT 
  cout << "SteadyStateAnalyser (" << this << "): dynamic SOR complete (" << timeUsed << " CPU seconds)." << endl;
  #endif
}*/
/*
void SteadyStateAnalyser::gaussSeidel()
{
  Vector<double> solution, last;
  int k,m;
  double diffNorm, solNorm, solInfNorm, convergence;

  cout << "AFTER MAKE Q: " << endl;
//  Q->print();
  solution.setLength(tangible);
  last.setLength(tangible);


  double omega = 1.2;
  double bestOmega = 0;
  long smallestIterations = 100000;
  long iterations;
  while (omega < 1.25) {
  for (long n=0; n<tangible; n++) {
    solution[n] = 1;
    last[n] = 1;
  }
//  cout << "OMEGA: " << omega << endl;
  iterations = 0;
  do {
    m = 1;
//    m = steps(iterations);
    for (k=0; k<m; k++) 
      Q->SORIteration(solution, omega);
    Q->calcResidualNorm(solution, solNorm, solInfNorm, last, diffNorm);
    iterations += m;
    cout << "iteration " << iterations <<" result's infnorm is " << solInfNorm << endl;
    if (diffNorm != 0) convergence = diffNorm / solNorm;
      else convergence = 0;
    for (n=0; n<tangible; n++) {
      last(n) = solution(n);
    }
  } while (convergence > 10*EPS && iterations < 200);

  if (iterations < smallestIterations) {
    smallestIterations = iterations;
    bestOmega = omega;
  }
  cout << "OMEGA: " << omega << " ITERATIONS: " << iterations << endl;
  omega += 0.0005; //0.0025;
  }
  cout << "BEST OMEGA: " << bestOmega << " smallest Its: " << smallestIterations << endl; 
  solution.normalise();
  #ifndef RELEASE
  solution.textOutput("output.sol");
  #endif
  unScale(solution);
  solution.normalise();
  for (long n=0; n<tangible; n++) {
    cout << "final[" << n << "] = " << setprecision(20) << solution(n) << endl;
  }
  cout << "result's infNorm: " << solInfNorm << endl;
  solution.output(SteadyVectorFile);
}

void SteadyStateAnalyser::gaussSeidel()
{
  Vector<double> solution, last;
  int k,m;
  double diffNorm, solNorm, solInfNorm, convergence;

  cout << "AFTER MAKE Q: " << endl;
//  Q->print();
  solution.setLength(tangible);
  last.setLength(tangible);


  double omega = 1.0;
  double bestOmega = 0;
  long smallestIterations = 100000;
  double smallestNorm = 500;
  long iterations;
  for (long n=0; n<tangible; n++) {
    solution[n] = 1;
  }
  for(n=0; n<20; n++) {
    Q->SORIteration(solution,1.0);
  }
  for ( n=0; n<tangible; n++) {
    last[n] = solution(n);
  }
  while (omega < 1.5) {
  for (long n=0; n<tangible; n++) {
    solution[n] = last(n);
  }
//  cout << "OMEGA: " << omega << endl;
  iterations = 0;
  for (k=0; k<1; k++)
    Q->SORIteration(solution, omega);
  Q->calcResidualNorm(solution, solNorm, solInfNorm, last, diffNorm);
  cout << "OMEGA: " << omega <<"  NORM: " << log10(solInfNorm) << endl;

  if (solInfNorm < smallestNorm) {
    smallestNorm = solInfNorm;
    bestOmega = omega;
  }
  omega += 0.0025;
  }
  cout << "BEST OMEGA: " << bestOmega << " smallest norm: " << smallestNorm << endl; 
  solution.normalise();
  #ifndef RELEASE
  solution.textOutput("output.sol");
  #endif
  unScale(solution);
  solution.normalise();
//  for (n=0; n<tangible; n++) {
//    cout << "final[" << n << "] = " << setprecision(20) << solution(n) << endl;
//  }
  cout << "result's infNorm: " << solInfNorm << endl;
  solution.output(steadyVectorFile);
}
*/

