////////////////////////////////////////////////////////////////////
// WMCA Markov Chain Analyser        (C) 1995 William Knottenbelt //
////////////////////////////////////////////////////////////////////
//            Data Network Architectures Laboratory               //
//               Department of Computer Science                   //
//                   University of Cape Town                      //
//                       Rondebosch 7700                          //
//                        SOUTH AFRICA                            //
////////////////////////////////////////////////////////////////////
// email: william@cs.uct.ac.za             phone: +27 21 650 2670 //
////////////////////////////////////////////////////////////////////

using namespace std;

#include <iostream>
#include <cassert>
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <complex>
#include <cstring>
#include "common.hpp"
#include "dyna.hpp"
#include "store.hpp"
#include "avl.hpp"
#include "info.hpp"
#include "element.hpp"
#include "vector.hpp"
#include "sparse-nocomplex.hpp"

#ifndef LOGGING
#define SILENT 1
#endif


long SparseMatrixRow::addElement(long column, double *index) {
#ifdef SAFE
  //  cout << "column = " << column << endl;
  //  cout << "index = " << index << " (" << *index << ")" << endl;
  //  cout << "entries = " << entries << endl;
  for (long n=0; n<entries; n++)  
    if (element(n).getColumn() == column) {
      cout << "SMR::addElement!! col = " << column << " index = " << index << "(entries = " << entries << ")" << endl;
//      assert(0);
//      element(n).setEntry(element(n).getEntry()+entry);
//      return n;
    }
#endif
  element.check(entries);
  element[entries].setColumn(column);
  element[entries].setIndex(index);
  return entries++;
}

long SparseMatrixRow::safeAddElement(long column, double *index) {
  for (long n=0; n<entries; n++)  
    if (element(n).getColumn() == column) {
      element(n).setIndex(index);
      return 0;
    }
  element.check(entries);
  element[entries].setColumn(column);
  element[entries].setIndex(index);
  entries++;
  return 1;
}


void SparseMatrixWithRowColumn::multiply(Vector<double> &x, Vector <double> &result) const {
 register long n,t;
  register double sum;
  register SparseMatrixRow *_row;

  for (n=0; n<rows; n++) {
    sum = 0; // ### think about extra var here for entries (don't)
    _row = &row(n);
    for (t=0; t<_row->getEntries(); t++) {
      sum +=
      (*(_row->getIndex(t))) * x(_row->getColumn(t));
    }
    result[n] = sum;
  }
//  cout << "total states: " << total << "\n";
  #ifdef VERBOSE
  cout << "multiply done" << endl;
  #endif
}

void SparseMatrix::transMultiply(Vector<double> &x, Vector <double> &result) const {
  register double sum;
  register long n, t;
  register SparseMatrixCol *_col;

  for (n=0; n<cols; n++) {
    _col = &col(n);
    sum = 0;
    for (t=0; t<_col->getEntries(); t++) {
      sum += (*(_col->getIndex(t)))*x(_col->getRow(t));
    }
    result[n] = sum;
  }
}

void SparseMatrixWithRow::multiply(Vector<double> &x, Vector <double> &result) const {
  register double sum;
  register long n, t;
  register SparseMatrixRow *_row;

#ifdef SAFE
  assert(result.getLength() == rows);
#endif
  for (n=0; n<rows; n++) {
    _row = &row(n);
    sum = 0;
    for (t=0; t<_row->getEntries(); t++) {
#ifdef SAFE
      assert( _row->getColumn(t) >= 0 && _row->getColumn(t) < x.getLength());
#endif
      sum += (*(_row->getIndex(t)))*x(_row->getColumn(t));
    }
    result[n] = sum;
  }
}

void SparseMatrixWithRow::integrity(char *message) const {
  register long n, t;
  register SparseMatrixRow *_row;

  for (n=0; n<rows; n++) {
    _row = &row(n);
    assert(_row);
    for (t=0; t<_row->getEntries(); t++) {
      if (_row->getColumn(t) < 0 || _row->getColumn(t) >= rows)
	cout << "!!! WARNING (" << message <<") !!! row "<< n << " entry " << t << " is " << _row->getColumn(t) << " should be in range (0, " << rows << ")" << endl;

      assert( _row->getColumn(t) >= 0 && _row->getColumn(t) < rows);
    }
  }
  cout << "&&&& integrity check (" << message << ") on matrix ok &&&&" << endl;
}


void SparseMatrixWithRow::inner(Vector<double> &x, Vector <double> &fixed) const {
  register double sum;
  register long n, t;
  register SparseMatrixRow *_row;

#ifdef SAFE
  //  assert(result.getLength() == rows);
#endif
  for (n=0; n<rows; n++) {
    //    if(n%1000==0)
      //      cout << "************ row = " << n << endl;
    _row = &row(n);
    assert(_row);
    sum = fixed[n];
    for (t=0; t<_row->getEntries(); t++) {
      //#ifdef SAFE

      if (_row->getColumn(t) < 0 || _row->getColumn(t) >= x.getLength()) 
	cout << "!!! WARNING !!! row "<< n << " entry " << t << " is " << _row->getColumn(t) << " should be in range (0, " << x.getLength() << ")" << endl;

      assert( _row->getColumn(t) >= 0 && _row->getColumn(t) < x.getLength());
      //#endif
      sum += (*(_row->getIndex(t)))*x(_row->getColumn(t));
    }
    x[n] = sum;
  }
  //  cout << "&&&&&&&&&&&&&&&&&& all done &&&&&&&&&&&&&&&&&&&" << endl;
}

void SparseMatrixWithRow::inner(Vector<double> &x, Vector <double> &fixed, double &conv) const {
  register double sum;
  register long n, t;
  register SparseMatrixRow *_row;

#ifdef SAFE
  //  assert(result.getLength() == rows);
#endif

  double maxSol = 0, maxDiff = 0;

  for (n=0; n<rows; n++) {
    _row = &row(n);
    sum = fixed[n];
    for (t=0; t<_row->getEntries(); t++) {
#ifdef SAFE
      assert( _row->getColumn(t) >= 0 && _row->getColumn(t) < x.getLength());
#endif
      sum += (*(_row->getIndex(t)))*x(_row->getColumn(t));
    }
    if (fabs(sum) > maxSol)
      maxSol = fabs(sum);
    if (fabs(sum - x[n]) > maxDiff)
      maxDiff = fabs(sum - x[n]);
    x[n] = sum;
  }
  conv = maxDiff/maxSol;
}

void SparseMatrixWithRow::innerSOR(Vector<double> &x, Vector <double> &fixed, int lower, int upper, double omega) const {
  register double sum;
  register long n, t;
  register SparseMatrixRow *_row;

  //  double omega = 1.2;
  double x_new = 0.0;

  if(upper > rows)
    upper = rows;

#ifdef SAFE
  //  assert(result.getLength() == rows);
#endif
  for (n=lower; n<upper; n++) {
    //    if(n%1000==0)
      //      cout << "************ row = " << n << endl;
    _row = &row(n);
    assert(_row);
    sum = fixed[n];
    for (t=0; t<_row->getEntries(); t++) {
      //#ifdef SAFE

      if (_row->getColumn(t) < 0 || _row->getColumn(t) >= x.getLength()) 
	cout << "!!! WARNING !!! row "<< n << " entry " << t << " is " << _row->getColumn(t) << " should be in range (0, " << x.getLength() << ")" << endl;

      assert( _row->getColumn(t) >= 0 && _row->getColumn(t) < x.getLength());
      //#endif
      sum += (*(_row->getIndex(t)))*x(_row->getColumn(t));
    }
    //    x[n] = sum;
    x_new = sum;
    x[n] = omega * x_new + ((1.0-omega)*x[n]);
  }
  //  cout << "&&&&&&&&&&&&&&&&&& all done &&&&&&&&&&&&&&&&&&&" << endl;
}


void SparseMatrixWithRow::innerSOR(Vector<double> &x, Vector <double> &fixed, double &maxSol, double &maxDiff, int lower, int upper, double omega) const {
  register double sum;
  register long n, t;
  register SparseMatrixRow *_row;

  //  double omega = 1.2;
  double x_new = 0.0;


  if(upper > rows)
    upper = rows;

#ifdef SAFE
  //  assert(result.getLength() == rows);
#endif

  //  double maxSol = 0, maxDiff = 0;

  for (n=lower; n<upper; n++) {
    _row = &row(n);
    sum = fixed[n];
    for (t=0; t<_row->getEntries(); t++) {
#ifdef SAFE
      assert( _row->getColumn(t) >= 0 && _row->getColumn(t) < x.getLength());
#endif
      sum += (*(_row->getIndex(t)))*x(_row->getColumn(t));
    }
    if (fabs(sum) > maxSol)
      maxSol = fabs(sum);
    if (fabs(sum - x[n]) > maxDiff)
      maxDiff = fabs(sum - x[n]);
    //    x[n] = sum;    
    x_new = sum;
    x[n] = omega * x_new + ((1.0-omega)*x[n]);
  }
  //  conv = maxDiff/maxSol;
}
void SparseMatrixWithRow::innerSOR(Vector<double> &x, Vector <double> &fixed, double omega) const {
  register double sum;
  register long n, t;
  register SparseMatrixRow *_row;

  //  double omega = 1.2;
  double x_new = 0.0;


#ifdef SAFE
  //  assert(result.getLength() == rows);
#endif
  for (n=0; n<rows; n++) {
    //    if(n%1000==0)
      //      cout << "************ row = " << n << endl;
    _row = &row(n);
    assert(_row);
    sum = fixed[n];
    for (t=0; t<_row->getEntries(); t++) {
      //#ifdef SAFE

      if (_row->getColumn(t) < 0 || _row->getColumn(t) >= x.getLength()) 
	cout << "!!! WARNING !!! row "<< n << " entry " << t << " is " << _row->getColumn(t) << " should be in range (0, " << x.getLength() << ")" << endl;

      assert( _row->getColumn(t) >= 0 && _row->getColumn(t) < x.getLength());
      //#endif
      sum += (*(_row->getIndex(t)))*x(_row->getColumn(t));
    }
    //    x[n] = sum;
    x_new = sum;
    x[n] = omega * x_new + ((1.0-omega)*x[n]);
  }
  //  cout << "&&&&&&&&&&&&&&&&&& all done &&&&&&&&&&&&&&&&&&&" << endl;
}


void SparseMatrixWithRow::innerSOR(Vector<double> &x, Vector <double> &fixed, double &conv, double omega) const {
  register double sum;
  register long n, t;
  register SparseMatrixRow *_row;

  //  double omega = 1.2;
  double x_new = 0.0;


#ifdef SAFE
  //  assert(result.getLength() == rows);
#endif

  double maxSol = 0, maxDiff = 0;

  for (n=0; n<rows; n++) {
    _row = &row(n);
    sum = fixed[n];
    for (t=0; t<_row->getEntries(); t++) {
#ifdef SAFE
      assert( _row->getColumn(t) >= 0 && _row->getColumn(t) < x.getLength());
#endif
      sum += (*(_row->getIndex(t)))*x(_row->getColumn(t));
    }
    if (fabs(sum) > maxSol)
      maxSol = fabs(sum);
    if (fabs(sum - x[n]) > maxDiff)
      maxDiff = fabs(sum - x[n]);
    //    x[n] = sum;    
    x_new = sum;
    x[n] = omega * x_new + ((1.0-omega)*x[n]);
  }
  conv = maxDiff/maxSol;
}

void SparseMatrix::calcResidual(Vector<double> &x, Vector <double> &result) const {
  register double sum;
  register long n,t;  
  register SparseMatrixCol *_col;

  for (n=0; n<cols; n++) {
    _col = &col(n);
    sum = 0;
    for (t=0; t<_col->getEntries(); t++) {
      sum += (*(_col->getIndex(t)))*x(_col->getRow(t));
    }
    result[n] = -sum;
  }
}

void SparseMatrix::calcResidualNorm(Vector<double> &x, double &xNorm, double &xInfNorm, Vector<double> &last, double &xDiffNorm) const
{
  register double sum;
  
  register long n,t,r;
  register double *index;
  register double weight;
  register SparseMatrixCol *_col;

  xNorm = xInfNorm = 0;
  xDiffNorm = 0;
  for (n=0; n<cols; n++) {
    _col = &col(n);
    sum = 0;
    for (t=0; t<_col->getEntries(); t++) {
      r = _col->getRow(t);
      index = _col->getIndex(t);
      weight = *index;
      sum += weight*x(r);
    }
    if (fabs(x(n) - last(n)) > xDiffNorm) xDiffNorm = fabs(x(n) - last(n));
    if (fabs(x(n)) > xNorm) xNorm = fabs(x(n));
//    xInfNorm += sum*sum;
    if (fabs(sum) > xInfNorm) xInfNorm = fabs(sum);
  }
//  xInfNorm = sqrt(xInfNorm);
}

void SparseMatrix::calcResidualNormAI(Vector<double> &x, double &xNorm, double &xInfNorm, Vector<double> &last, double &xDiffNorm) const
{
  register long n,t,r;
  register double *index;
  register double sum, weight;
  register SparseMatrixCol *_col;

  xNorm = 0;
  xInfNorm = 0;
  xDiffNorm = 0;
  for (n=0; n<cols; n++) {
    sum = 0;
    _col = &col(n);
    for (t=0; t<_col->getEntries(); t++) {
      r = _col->getRow(t);
      index = _col->getIndex(t);
      weight = *index;//store->getEntry(index);
      sum += weight*x(r);
    }
    if (fabs(x(n) - last(n)) > xDiffNorm) xDiffNorm = fabs(x(n) - last(n));
    if (fabs(x(n)) > xNorm) xNorm = fabs(x(n));
    if (fabs(sum - x(n)) > xInfNorm) xInfNorm = fabs(sum - x(n));
    last[n] = x(n);
  }
}

void SparseMatrix::twinCalcResidualNorm(Vector<double> &x1, double &x1Norm, double &x1InfNorm, Vector<double> &x2, double &x2Norm, double &x2InfNorm, Vector<double> &last, double &x1DiffNorm, double &x2DiffNorm) const
{
  register double sum1, sum2;
  
  register long n,t,r;
  register double weight;
  register SparseMatrixCol *_col;

  x1Norm = x2Norm = x1InfNorm = x2InfNorm = 0;
  x1DiffNorm = 0;
  x2DiffNorm = 0;
  for (n=0; n<cols; n++) {
    _col = &col(n);
    sum1 = sum2 = 0; // ### think about an extra var here for entries
    for (t=0; t<_col->getEntries(); t++) {
      r = _col->getRow(t);
      weight = *(_col->getIndex(t));
      sum1 += weight*x1(r);
      sum2 += weight*x2(r);
    }
    if (fabs(x1(n) - last(n)) > x1DiffNorm) x1DiffNorm = fabs(x1(n) - last(n));
    if (fabs(x2(n) - last(n)) > x2DiffNorm) x2DiffNorm = fabs(x2(n) - last(n));
    sum1 = fabs(sum1);
    sum2 = fabs(sum2);
    if (fabs(x1(n)) > x1Norm) x1Norm = fabs(x1(n));
    if (fabs(x2(n)) > x2Norm) x2Norm = fabs(x2(n));
//    x1InfNorm += sum1*sum1;
//    x2InfNorm += sum2*sum2;
    if (sum1 > x1InfNorm) x1InfNorm = sum1;
    if (sum2 > x2InfNorm) x2InfNorm = sum2;
  }
//  x1InfNorm = sqrt(x1InfNorm);
//  x2InfNorm = sqrt(x2InfNorm);
}

void SparseMatrix::preMultiply(Vector<double> &x, Vector<double> &result) const
{
  transMultiply(x, result);
}


void SparseMatrix::SORIteration(Vector<double> &x, double omega) const {
  double sum;
  
  register long n, t, r;
  register SparseMatrixCol *_col;

//  long total = 0;

  for (n=0; n<cols; n++) {
    _col = &col(n);
    sum = 0; // ### think about an extra var here for entries
    for (t=0; t<_col->getEntries(); t++) {
      r = _col->getRow(t);
      if (r!=n) {
        sum += (*(_col->getIndex(t)))*x(r);
      } /*  else if ((weight-1)>EPS) {
        cout << "WHOA!!! r=" << r << ", n= " << n << " weight = " << weight << endl;
        assert(weight==1); 
      } */
//      total++;
    }
    x[n] = omega*sum + (1-omega)*x(n);
//    assert(x(n) >= 0);
//    cout << "x [" << n << "] = " << x[n] << endl;
  }
//  cout << "total states: " << total << "\n";
  #ifdef VERBOSE
  cout << "calcResidual done" << endl;
  #endif
}

void SparseMatrix::safeSORIteration(Vector<double> &x, double omega) const {
  double sum;
  
  long n,t,r;
  double pivot = 0 ;

  for (n=0; n<cols; n++) {
    sum = 0; // ### think about an extra var here for entries
    for (t=0; t<col(n).getEntries(); t++) {
      r = col(n).getRow(t);
      if (r != n) {
        sum += (*(col(n).getIndex(t)))*x(r);
      } else 
	pivot = *(col(n).getIndex(t));
//      total++;
    }
    assert (pivot != 0) ;
    x[n] = omega*sum/pivot + (1-omega)*x(n);
//    x[n] = sum/pivot;
//    assert(x(n) >= 0);
//    cout << "x [" << n << "] = " << x[n] << endl;
  }
  #ifdef VERBOSE
  cout << "calcResidual done" << endl;
  #endif
}

void SparseMatrix::twinSORIteration(Vector<double> &x1, double omega1, Vector<double> &x2, double omega2) const {

  register long n,t,r;
  register double sum;
  register SparseMatrixCol *_col;

// long total = 0;

  for (n=0; n<cols; n++) {
    _col = &col(n);
    sum = 0; // ### think about an extra var here for entries
//    entries = _col->getEntries();
    for (t=0; t<_col->getEntries(); t++) {
      r = _col->getRow(t);
      // ### can be made faster by putting r != n check here
      if (r != n)
        sum += (*(_col->getIndex(t)))*x1(r);
//      total++;
    }
    x1[n] = omega1*sum + (1.0-omega1)*x1(n);
    x2[n] = omega2*sum + (1.0-omega2)*x2(n);
//    assert(x(n) >= 0);
//    cout << "x [" << n << "] = " << x[n] << endl;
  }
//  cout << "total states: " << total << "\n";
  #ifdef VERBOSE
  cout << "twinSORIteration done" << endl;
  #endif
}

void SparseMatrix::print()
{
  long n,t;
  for (n=0; n<cols; n++) {    
    cout << "Col " << n << ": ";
    for (t=0; t<col(n).getEntries(); t++)
      cout << "(" << col(n).getRow(t) << "," << *(col(n).getIndex(t)) << ") ";
    cout << endl;
  }
}

void SparseMatrixWithRow::print()
{
  long n,t;
  for (n=0; n<rows; n++) {    
    cout << "Row " << n << ": ";
    for (t=0; t<row(n).getEntries(); t++)
      cout << "(" << row(n).getColumn(t) << "," << *(row(n).getIndex(t)) << ") ";
    cout << endl;
  }
}


void SparseMatrixWithRowColumn::print()
{
  long n,t;
  for (n=0; n<rows; n++) {    
    cout << "Row " << n << ": ";
    for (t=0; t<row(n).getEntries(); t++)
      cout << "(" << row(n).getColumn(t) << "," << *(row(n).getIndex(t)) << ") ";
    cout << endl;
  }
}

void DirectMatrix::print()
{
  long n,t;
  for (n=0; n<rows; n++) {    
    cout << "Row " << n << ": ";
    for (t=0; t<row(n).getEntries(); t++)
      cout << "(" << row(n).getColumn(t) << "," << *(row(n).getIndex(t)) << ") ";
    cout << endl;
  }
}

SparseMatrix::SparseMatrix(long r, long c)
{
  #ifndef SILENT
  cout << "SparseMatrix (" << this << "): constructing " << r << " rows and " << c << " columns...\n";
  #endif
  rows = r;
  cols = c;
  calls = 0;
  col.setLength(cols);
  store = new Store;
  tree = new AVLTree(store);
  #ifndef SILENT
  cout << "SparseMatrix (" << this << "): constructed with " << r << " rows and " << c << " columns." << endl;
  #endif
}

SparseMatrixWithRow::SparseMatrixWithRow(long r, long c)
{
  #ifdef VERBOSE
  cout << "SparseMatrixWithRow (" << this << "): constructing " << r << " rows and " << c << " columns...\n";
  #endif
  rows = r;
  cols = c;
  calls = 0;
  row.setLength(rows);
  store = new Store;
  tree = new AVLTree(store);
  #ifndef SILENT
  cout << "SparseMatrixWithRow (" << this << "): constructed with " << r << " rows and " << c << " columns." << endl;
  #endif
}

SparseMatrix::~SparseMatrix()
{
  #ifdef VERBOSE
  cout << "SparseMatrix (" << this << "): destructor...\n";
  #endif
  delete tree;
  delete store;
  #ifdef VERBOSE
  cout << "SparseMatrix (" << this << "): destroyed\n";
  #endif
}

SparseMatrixWithRow::~SparseMatrixWithRow()
{
  #ifdef VERBOSE
  cout << "SparseMatrixWithRow (" << this << "): destructor...\n";
  #endif
  delete tree;
  delete store;
  #ifdef VERBOSE
  cout << "SparseMatrixWithRow (" << this << "): destroyed\n";
  #endif
}

void SparseMatrix::statistics() {
  #ifndef SILENT
  cout << "SparseMatrix (" << this << ") statistics: " << calls << " non-zero elements used " << store->getElements() << " doubles." << endl;
  #endif
}


