////////////////////////////////////////////////////////////////////
// WMCA Markov Chain Analyser        (C) 1995 William Knottenbelt //
////////////////////////////////////////////////////////////////////
//            Data Network Architectures Laboratory               //
//               Department of Computer Science                   //
//                   University of Cape Town                      //
//                       Rondebosch 7700                          //
//                        SOUTH AFRICA                            //
////////////////////////////////////////////////////////////////////
// email: william@cs.uct.ac.za             phone: +27 21 650 2670 //
////////////////////////////////////////////////////////////////////

//#define SAFE

//#define DEBUG
//#define VERBOSE

#define HASH_LENGTH 350003
// 16384
#define BUFFERS 2
//#define BUFFER_SIZE 16384

using namespace std;

#include <iostream>
#include <iomanip>
#include <cassert>
#include <cstdarg>
#include <cmath>
#include <cstdio>
#include <cstdlib>
#include <complex>
#include "common.hpp"
#include "dyna.hpp"
#include "info.hpp"
#include "timer.hpp"
#include "element.hpp"
#include "user.hpp"
#include "hash.hpp"
#include "full.hpp"
#include "cluster.hpp"
#include "state.hpp"
#include "filenames.hpp"
#ifdef RESOURCE_USAGE
#include <unistd.h>
#endif


ostream &operator <<(ostream &out, const StateMatrixElement &source)
{
  out << "StateMatrixElement: destination [" << source.getDestination() << "]  weight [" << source.getWeight() << "]" << endl;
  return out;
}

ostream &operator <<(ostream &out, const HashTableEntry &source)
{
  out << "HashTableEntry: secondary [" << source.getSecondaryKey() << "]  target[" << source.getTarget() << "]\n";
  return out;
} 


StateHashTable::StateHashTable() {
  #ifdef LOGGING
  cout << "StateHashTable (" << this << "): default initialise\n";
  #endif
  rows.setLength(HASH_LENGTH);
  accesses = 0;
  maxRowLength = 0;
  #ifdef LOGGING
  cout << "StateHashTable (" << this << "): constructed\n";
  #endif
}

StateHashTable::~StateHashTable() {
  #ifdef LOGGING
  cout << "StateHashTable (" << this << "): destructor\n";
  #endif
  #ifdef LOGGING
  cout << "StateHashTable (" << this << "): destroyed\n";
  #endif
}

void StateHashTable::statistics()
{
#ifdef LOGGING
  int used=0;
  long totalLength=0;
  cout << "StateHashTable (" << this << ") statistics:\n";
  long n;
  for (n=0; n<HASH_LENGTH; n++)
    if (rows(n).getEntries()) {
      used++;
      totalLength += rows(n).getEntries();
    }
  //double prob = 1.0;
  //long average = (long) (totalLength/used+0.5);
  //for (n=0; n<average; n++)
  //  prob *= (double) ( (unsigned) SECONDARY-n)/( (unsigned) SECONDARY);
  //prob = pow(prob, HASH_LENGTH);
  cout << setw(11) << totalLength << " states stored in hash table." << endl;
  //cout << setw(11) << prob << " estimated probability of complete state space coverage." << endl;
  cout << setw(11) << accesses << " accesses to hash table." << endl;
  cout << setw(11) << used << " hash table rows used." << endl;
  cout << setw(11) << maxRowLength << " was the maximum hash table row length." << endl;
  if (used)
    cout << setw(11) << (double) totalLength/used << " was the average hash table row length." << endl;
#endif
}

Generator::Generator(char* basename)
{    
   char* optionsFileName = add_extension(basename, "OPTIONS") ;
#ifdef LOGGING
  cout << "Generator (" << this << "): start up!" << endl;
#endif
#ifdef RESOURCE_USAGE
  core = sbrk(0);
#endif
  vanishStack = NULL;
  tangibleStack = NULL;
  hashTable = NULL;
  timer = NULL;
  info = NULL;
  buffer[0] = buffer[1] = NULL;  
  info = new Information;
  timer = new Timer;
  for (int b=0; b<BUFFERS; b++)
    buffer[b] = new char[BUFFER_SIZE];

  hashTable = new StateHashTable;
  tangibleStack = new StateQueue;
  vanishStack = new ClusterStack;

  if (!info->input(optionsFileName)) 
     terminate("cannot read options file");

  maxStates = info->getMaxStates();
  maxCPUTime = info->getMaxCPUTime();
  watchClock = (maxCPUTime > 0.0);
  repInterval = info->getGenRepInterval();
  repStyle = info->getGenRepStyle();
  #ifdef LOGGING
  cout << "Generator (" << this << "): powered up!" << endl;
  #endif
}

Generator::~Generator() {
  destroy();
}

void Generator::destroy()
{
  #ifdef LOGGING
  cout << "Generator (" << this << "): destructor" << endl;
  #endif
  if (vanishStack) delete vanishStack;
  if (tangibleStack) delete tangibleStack;
  if (hashTable) delete hashTable;
  if (timer) delete timer;
  if (info) delete info;

  for (int b=0; b<BUFFERS; b++)
    if (buffer[b]) delete [] buffer[b];
  #ifdef LOGGING
  cout << "Generator (" << this << "): powered down!" << endl;
  #endif
}

int Generator::openFile(char *filename, FILE *&fp, char *buff)
{
  fp = fopen(filename, "w");
  if (!fp) terminate("state.cxx cannot open file '%s'", filename);
  if (setvbuf(fp, buff, _IOFBF, BUFFER_SIZE))
    terminate("cannot allocate buffer for file '%s'", filename);
  return 1;
}

int Generator::closeFile(FILE *fp)
{
  return fclose(fp);
}

void Generator::pushStart(FullState *current) {
  FullState *target;
  double sum, weight, probability;
  int succ, found;
  long id;

  #ifdef SAFE
  assert(current);
  #endif
  current->findEnabledTransitions();

  if (current->getType() == TANGIBLE) {
    current->computeHashKeys();
    pushTangible(current); // register
    return;
  }
  // vanishing state... whoa.
  pushVanishing(current, 1);
  //  FullState *next = new FullState;
  FullState *immediate = new FullState;
  while (vanishStack->getHeight()) {
    popVanishing(target, weight);
    #ifdef SAFE
    assert(target);
    assert(target->getType()==VANISHING);
    #endif
    sum = target->getTotalTransWeight();
    for (succ = 0; succ < target->getNumberEnabled(); succ++) {
      probability = target->getEnabledTransWeight(succ)/sum;
      target->fireEnabledTransition(succ, immediate);
      immediate->findEnabledTransitions();              
      switch(immediate->getType()) {
      case VANISHING:
        if (pushVanishing(immediate, probability*weight))
	  immediate = new FullState;
        break;
      case TANGIBLE:
        immediate->computeHashKeys();
        found = hashTable->searchState(immediate->getPrimaryKey(), immediate->getSecondaryKey(),id); 
        if (!found) {
          pushTangible(immediate);
          immediate = new FullState;
        }
      } // switch immediate
    } // for succ
    delete target;
  } // while vanish->stackHeight
  #ifdef DEBUG
  cout << "push start exit" << endl;
  #endif
}

void Generator::statistics() {
#ifdef LOGGING
  double timeUsed = timer->observe();
  hashTable->statistics();
  if (timeUsed != 0) {
    cout << "Generation speed statistics:" << endl;
    cout << setw(11) << timeUsed << " CPU seconds used." << endl;
    cout << setw(11) << tangible/timeUsed << " tangible states generated/CPU second." << endl;
    cout << setw(11) << hashTable->getAccesses()/timeUsed << " accesses/CPU second." << endl;
  }
#endif
#ifdef STATE_COUNT
  cout << setw(11) << FullState::instances << " FullState instances created." << endl << setw(11) << FullState::destroyed << " FullState instances destroyed." << endl;
#endif
#ifdef RESOURCE_USAGE
    unsigned long memory = (unsigned long) (sbrk(0) - core);
    cout << "Generator (" << this <<"): " << (double) (memory/1024.0) << " KB memory used for data." << endl;
#endif
}

int Generator::explore(FullState *current, char* basename)
{
  int trans, enabled, t, succ, found;
  long id, connections;
  FullState *next, *target;
  double weight, sum, probability;
  char* infoFileName =  add_extension(basename, "INFO") ;

  current->initialise(info->getTransitions(), info->getTransType(), info->getPriority());
  current->initial();
  info->setStateSize(current->size());

  timer->start();
  if (!openFile(info->getStateListFile(),sf,buffer[0])) return 1;
  if (!openFile(info->getTransMatrixFile(),mf,buffer[1])) return 1;
  unexplored = 0;
  maxChildren = 0;
  tangible = 0;

  /*
    We set trans to be 0, trans is reset within the while-loop
    but also used *after* the while-loop, the compiler thinks of
    course that it is possible to skip over the while-loop and hence
    for trans to be used uninitialised. Actually since pushStart
    will call pushTangible unexplored should be greater than 0 for
    at least the first test.
  */
  trans = 0 ;

  pushStart(current);

  next = new FullState;         
  FullState *immediate = new FullState;
  while(unexplored) {
    current = popTangible();
#ifdef DEBUG
    cout << "POPPED! (states " << tangible << ") ("  << current->getPrimaryKey() << "," << current->getSecondaryKey() << ")" << endl;
    current->output();
#endif
    #ifdef SAFE
    assert(current->getType()==TANGIBLE);
    #endif
    trans = current->getNumberEnabled();
#ifdef DEBUG
    cout << "we have a tangible marking with " << trans << " enabled transitions\n";
#endif
    if (!trans) deadlock(current);
    connections = 0;
    for (t = 0; t < trans; t++) {
#ifdef DEBUG
      cout << "- now firing transition " << t << endl;
#endif
      current->fireEnabledTransition(t, next);
      next->computeHashKeys();
      found = hashTable->searchState(next->getPrimaryKey(), next->getSecondaryKey(),id);
      if (found) { // found a tangible state
#ifdef DEBUG
        cout << "* firing to tangible that's already there (" << id << ") !\n";
        cout << "* connections: " << connections << endl;        
#endif
        current->setChild(t, id, current->getEnabledTransWeight(t));
        connections++;
#ifdef DEBUG
        cout << "* left\n";
#endif
      } else {
        next->findEnabledTransitions();
        weight = current->getEnabledTransWeight(t);
        switch(next->getType()) {
	case TANGIBLE:
          current->setChild(t, pushTangible(next), weight);
          next = new FullState;
          connections++;
	  break;
	case VANISHING:
          if (pushVanishing(next, weight))
	    next = new FullState;
          while (vanishStack->getHeight()) {
            popVanishing(target, weight);
            // find successors and probs of this
            sum = target->getTotalTransWeight();
#ifdef DEBUG
            cout << "# " << target->getNumberEnabled() << " kids have a total weight of " << sum << endl;
#endif
            enabled = target->getNumberEnabled();
            if (!enabled) deadlock(target);
	    for (succ = 0; succ < enabled; succ++) {
              probability = target->getEnabledTransWeight(succ)/sum;
#ifdef DEBUG
              cout << "# now attending to kid " << succ << "(probability " << probability << ")" << endl;
#endif
              // find the successor
              target->fireEnabledTransition(succ, immediate);
              immediate->computeHashKeys();
              found = hashTable->searchState(immediate->getPrimaryKey(), immediate->getSecondaryKey(), id);
              if (found) {
                current->setChild(t, id, probability*weight);
                connections++;
	      } else {
                immediate->findEnabledTransitions();
                switch(immediate->getType()) {
	        case VANISHING:
                  if (pushVanishing(immediate, probability*weight))
		    immediate = new FullState;
                  break;
 	        case TANGIBLE:
                  current->setChild(t, pushTangible(immediate), probability*weight);
                  immediate = new FullState;
                  connections++;
		  break;
	        } // switch immediate
	      } // if (found)... else
	    } // for succ
            delete target;
	  } // while vanish->stackHeight
	  break;
	} // switch next->getType
      }
    } // for t=0; t<trans
    if (connections > maxChildren) maxChildren = connections;
#ifdef DEBUG
      cout << "ID: " << id << "\n";
    cout << "state: " << tangible << " connections: " << connections << endl;
#endif
    current->matrixOutput(mf);
    current->output(sf);
    delete current;  
  } // while unexplored
  closeFile(sf);
  closeFile(mf);
  info->setTangible(tangible);
  info->setTrans(trans);
  info->setMaxChildren(maxChildren);
//  cout << "maxChildren: " << maxChildren << endl;
#ifdef LOGGING
  cout << "Generator (" << this << "): explore complete! (" << tangible << " tangible states)" << endl;
#endif
  statistics();
  info->setGeneration(1);
  assert(info->getGeneration() == 1);
  if (!info->outputStateInfo(infoFileName)) 
     terminate("cannot update state information file");
  return 1;
}

void Generator::terminate(char *fmt, ...)
{
  char buffer[2048];

  va_list argp;
  va_start(argp, fmt);
  vsprintf(buffer, fmt, argp); 
  va_end(argp);
  destroy();
  printf("\nfatal error: %s.\nexecution halted.\n", buffer);  
  exit(1);
}

int main(int argc, char *argv[])
{
   if (argc != 2)
   {
      printf("Must supply the basename as an argument\n") ;
      printf("Number of arguments was: %d\n", argc) ;
      exit(1) ;
   }
   
   Generator *jenny = new Generator(argv[1]);

   jenny->explore(new FullState, argv[1]);
   delete jenny;
   return 0;
}


