////////////////////////////////////////////////////////////////////
// WMCA Markov Chain Analyser        (C) 1995 William Knottenbelt //
////////////////////////////////////////////////////////////////////
//            Data Network Architectures Laboratory               //
//               Department of Computer Science                   //
//                   University of Cape Town                      //
//                       Rondebosch 7700                          //
//                        SOUTH AFRICA                            //
////////////////////////////////////////////////////////////////////
// email: william@cs.uct.ac.za             phone: +27 21 650 2670 //
////////////////////////////////////////////////////////////////////

using namespace std;

#include <iostream>
#include <fstream>
#include <string>
#include <cctype>
#include <cstdio>
#include <cassert>
#include <cstdarg>
#include <unistd.h>
#include "common.hpp"
#include "dyna.hpp"
#include "info.hpp"
#include "token.hpp"
#include "parse.hpp"

#define UNDEFINED -1

static KeywordEntry table[] = {
  { T_MODEL_DESCRIPTION,            "\\model"              },
  { T_STATE_VECTOR,                 "\\statevector"        },
  { T_TYPE,                         "\\type"               },
  { T_CONSTRUCTOR,                  "\\stateconstructor"   },
  { T_DESTRUCTOR,                   "\\statedestructor"    },
  { T_HEADER,                       "\\header"             },
  { T_TRANSITION,                   "\\transition"         },
  { T_INITIAL,                      "\\initial"            },
  { T_CONDITION,                    "\\condition"          },
  { T_ACTION,                       "\\action"             },
  { T_RATE,                         "\\rate"               },
  { T_WEIGHT,                       "\\weight"             },
  { T_PRIORITY,                     "\\priority"           },
  { T_GENERATION_CONTROL,           "\\generation"         },
  { T_MAX_STATES,                   "\\maxstates"          },
  { T_MAX_CPU_TIME,                 "\\maxcputime"         },
  { T_SAFE_HASH,                    "\\safehash"           },
  { T_REPORT_STYLE,                 "\\reportstyle"        },
  { T_REPORT_INTERVAL,              "\\reportinterval"     },
  { T_CONSTANT,                     "\\constant"           },
  { T_HELP_VALUE,                   "\\helpvalue"          },
  { T_PRIMARY,                      "\\primaryhash"        },
  { T_SECONDARY,                    "\\secondaryhash"      },
  { T_INVARIANT,                    "\\invariant"          },
  { T_SOLUTION_CONTROL,             "\\solution"           },
  { T_METHOD,                       "\\method"             },
  { T_ACCURACY,                     "\\accuracy"           },
  { T_MAX_ITERATIONS,               "\\maxiterations"      },
  { T_RELAX_PARAMETER,              "\\relaxparameter"     },
  { T_START_VECTOR,                 "\\startvector"        },
  { T_PERFORMANCE,                  "\\performance"        },
  { T_STATE_MEASURE,                "\\statemeasure"       },
  { T_COUNT_MEASURE,                "\\countmeasure"       },
  { T_ESTIMATOR,                    "\\estimator"          },
  { T_EXPRESSION,                   "\\expression"         },
  { T_PRECONDITION,                 "\\precondition"       },
  { T_POSTCONDITION,                "\\postcondition"      },
  { T_OUTPUT,                       "\\output"             },
  { T_STATELIST,                    "\\statelist"          },
  { T_STEADY_STATE_VECTOR,          "\\steadystatevector"  },
  { T_TRANSITION_MATRIX,            "\\transitionmatrix"   },
  { T_PERFORMANCE_RESULTS,          "\\performanceresults" },
  { T_PASSAGE,                      "\\passage"            },
  { T_TARGET_CONDITION,             "\\targetcondition"    },
  { T_SOURCE_CONDITION,             "\\sourcecondition"    },
  { T_SOJOURNTIME_LT,               "\\sojourntimeLT"      },
  { T_SOJOURNTIME,                  "\\sojourntimedist"    },
  { T_START,                        "\\t_start"            },
  { T_STOP,                         "\\t_stop"             },
  { T_STEP,                         "\\t_step"             },
  { T_EXCLUDE_CONDITION,            "\\exclude"            },
  { T_TRANSIENT,                    "\\transient"          },
  { T_GAUSS,                        "gauss"                },
  { T_LU,                           "lu"                   },
  { T_GRASSMANN,                    "grassmann"            },
  { T_POWER,                        "power"                },
  { T_GAUSS_SEIDEL,                 "gauss_seidel"         },
  { T_SOR,                          "sor"                  },
  { T_CGNR,                         "cgnr"                 },
  { T_CGS,                          "cgs"                  },
  { T_BICG,                         "bicg"                 },
  { T_BICGSTAB,                     "bicgstab"             },
  { T_BICGSTAB2,                    "bicgstab2"            },
  { T_MYCG,                         "mycg"                 },
  { T_AI,                           "ai"                   },
  { T_AIR,                          "air"                  },
  { T_AUTOMATIC,                    "automatic"            },
  { T_ON,                           "on"                   },
  { T_OFF,                          "off"                  },
  { T_FULL,                         "full"                 },
  { T_BRIEF,                        "brief"                },
  { T_NONE,                         "none"                 },
  { T_DYNAMIC,                      "dynamic"              },
  { T_DIAG,                         "diag"                 },
  { T_ILU,                          "ilu"                  },
  { T_MEAN,                         "mean"                 },
  { T_VARIANCE,                     "variance"             },
  { T_STDDEV,                       "stddev"               },
  { T_DISTRIBUTION,                 "distribution"         },
  { T_ALL,                          "all"                  },
  { T_UNDEFINED,                    "<undefined>"          }
};

static char *tokenStrings[] =  {

  // Keywords

  "T_MODEL_DESCRIPTION",     "T_STATE_VECTOR",          "T_TYPE",
  "T_CONSTRUCTOR",           "T_DESTRUCTOR",
  "T_HEADER",                "T_TRANSITION",
  "T_INITIAL",               "T_CONDITION",             "T_ACTION",
  "T_RATE",                  "T_WEIGHT",
  "T_PRIORITY",              "T_GENERATION_CONTROL",    "T_MAX_STATES",
  "T_MAX_CPU_TIME",          "T_SAFE_HASH",             "T_REPORT_STYLE",
  "T_REPORT_INTERVAL",       "T_CONSTANT",              "T_HELP_VALUE",
  "T_PRIMARY",               "T_SECONDARY",             "T_INVARIANT",
  "T_SOLUTION_CONTROL",      "T_METHOD",                "T_ACCURACY",
  "T_MAX_ITERATIONS",        "T_RELAX_PARAMETER",      
  "T_START_VECTOR",          "T_PERFORMANCE",           "T_STATE_MEASURE",
  "T_COUNT_MEASURE",         "T_ESTIMATOR",             "T_EXPRESSION",
  "T_PRECONDITION",          "T_POSTCONDITION",         "T_OUTPUT",
  "T_STATELIST",             "T_STEADY_STATE_VECTOR",   "T_TRANSITION_MATRIX",
  "T_PERFORMANCE_RESULTS",   "T_SOJOURNTIME_LT",        "T_START",
  "T_STOP",                  "T_STEP",                  "T_SOJOURNTIME",
  "T_PASSAGE",               "T_TRANSIENT",

  // Parameters

  "T_GAUSS",                 "T_LU",                    "T_GRASSMANN",
  "T_POWER", "T_GAUSS_SEIDEL",          "T_SOR",                   "T_CGNR",
  "T_CGS",
  "T_BICG",                  "T_BICGSTAB",              "T_BICGSTAB2",
  "T_MYCG", 
  "T_AI",                    "T_AIR",
  "T_AUTOMATIC",             "T_ON", 
  "T_OFF",                   "T_FULL",                  "T_BRIEF",
  "T_NONE",                  "T_DYNAMIC",
  "T_DIAG",                  "T_ILU",                   "T_MEAN",
  "T_VARIANCE",              "T_STDDEV",                "T_DISTRIBUTION",
  "T_ALL",    

  // Identifiers
  "<identifier>",

  // punctuation
  "{", "}", ",", ";", "%", "=", "<operator>",
  "->",

  // other
  "<end of file>", "<undefined>"
};

ostream &operator <<(ostream &out, const CodeBlock &source)
{
  if (source.getArray()) out << source.getArray();
  return out;
}

Tokeniser::Tokeniser()
{
  errors = 0;
  warnings = 0;
  suppressComment = 0;
  current = T_UNDEFINED;
}

void Tokeniser::format(char *message) 
{
  static int column = 0, width = CRT_WIDTH - 2;
  int start_word, stop_word, length, m, n;
  start_word = 0;
  if (!message || !message[0]) return;

  m = 0;
  do {
    if (message[m] == ' ' || message[m] == '\0' || message[m] == '\n') {
      stop_word = m;
      length = stop_word - start_word;
      if (column + length > width) {
        cout << endl << " ";
        column = 1;
      }
      for (n=start_word; n < stop_word; n++)
        cout << message[n];
      column += length;
      start_word = stop_word;
      if (message[m] == '\n') {
        column = 0;
      }
    }
  } while (message[m++]);
}

int Tokeniser::warning(const char *fmt, ...)
{
  char buffer[2048], message[2048];

  va_list argp;
  va_start(argp, fmt);
  vsprintf(buffer, fmt, argp); 
  va_end(argp);
  sprintf(message, "%s: %ld: (warning) %s.\n", sourceFile, lineNumber, buffer); 
  format(message);
  warnings++;
  return 0;
}

int Tokeniser::error(const char *fmt, ...)
{
  char buffer[2048], message[2048];

  va_list argp;
  va_start(argp, fmt);
  vsprintf(buffer, fmt, argp); 
  va_end(argp);
  sprintf(message, "%s: %ld: %s.\n", sourceFile, lineNumber, buffer);  
  format(message);
  errors++;
  if (errors > 5) {
    report();
    exit(1);
  }
  return 0;
}

int Tokeniser::openSource(char *filename)
{
  strcpy(sourceFile, filename);
  source.open(filename, ios::in);
  if (!source) return 0;
  source.unsetf(ios::skipws);
  lineNumber = 1;
  errors = 0;
  return nextChar();
}

int Tokeniser::nextChar() 
{
  source.get(ch);
  if (ch == '%' && !suppressComment)
    while (!source.eof() && ch != 10 && ch != 13) source.get(ch);

  if (source.eof()) ch = END_OF_FILE;
  
  if (ch==10 || ch==13) {
    ch = '\n';
    lineNumber++;
  }
  #ifdef DEBUG
  cout << ch;
  #endif
  return 1;
}

Token Tokeniser::scanIdentifier()
{
  int length = 0;
  suppressComment = 1;
  while (isalnum(ch) || ch=='.' || ch == '_' || (ch == '\\' && !length)) {
    identifier[length++] = ch;
    if (length==MAX_LENGTH) {
      error("Identifier exceeds maximum length (%d)", MAX_LENGTH);
      identifier[length - 1] = '\0';
      while (isalnum(ch) || ch=='.' || ch =='_') nextChar();
      suppressComment = 0;
      return T_IDENTIFIER;
    }
    nextChar();
  }
  identifier[length] = '\0';
  suppressComment = 0;
  #ifdef DEBUG
  cout << "scanned identifier: [" << identifier << "]\n";
  #endif

  for (int n=0; table[n].token != T_UNDEFINED ; n++)
    if (!strcmp(table[n].keyword,identifier)) {
      #ifdef DEBUG
      cout << "found keyword " << table[n].keyword << "\n";
      #endif
      return table[n].token;
    }
  return T_IDENTIFIER;
}

int Tokeniser::checkInteger()
{
  for (unsigned long n=0; n<strlen(identifier); n++)
    if (!isdigit(identifier[n])) return 0;
  return 1;
}

Token Tokeniser::scanReal()
{
  int length = 0;
  while (isalnum(ch) || ch == '.' || ch == '-' || (toupper(ch) == 'E' && !length)) {
    identifier[length++] = ch;
    if (length==MAX_LENGTH)
      error("Real number exceeds maximum length (%d)", MAX_LENGTH);
    nextChar();
  }
  identifier[length] = '\0';
  #ifdef DEBUG
  cout << "scanned real: [" << identifier << "]\n";
  #endif

  for (int n=0; table[n].token != T_UNDEFINED ; n++)
    if (!strcmp(table[n].keyword,identifier)) {
      #ifdef DEBUG
      cout << "found keyword " << table[n].keyword << "\n";
      #endif
      return table[n].token;
    }
  return T_IDENTIFIER;

}

Token Tokeniser::getNextToken()
{
  Token token = T_UNDEFINED;
  
  do {
    switch(ch) {
    case END_OF_FILE:
      return T_EOF;
    case '\n': 
      break;
    case ' ':
    case '\t':
      break;
    case '{':
      token = T_OPEN_BRACE;
      break;
    case '}':
      token = T_CLOSE_BRACE;
      break;
    case '%':
      token = T_PERCENT;
      break;
    case '=':
      token = T_EQUAL;
      break;
    case ',':
      token = T_COMMA;
      break;
    case ';':
      token = T_SEMICOLON;
      break;
    case '-':
      nextChar();
      if (ch == '>')
        token = T_DEREFERENCE;
      else {
        token = T_OPERATOR;
        return token;
      }
    case '+':
    case '/':
    case '*':
      token = T_OPERATOR;
      break;
    case '>':
    case '<':
      nextChar();
      token = T_OPERATOR;
      if (ch != '=') return token;
    default:
      if (isalnum(ch) || ch == '_' || ch == '\\')
        return scanIdentifier();
      else
        error("Unexpected character '%c' (%d)\n",ch,int(ch));
      break;
    }
    nextChar();
  } while (token == T_UNDEFINED);
  current = token;
  return token;
}

int Tokeniser::find(Token target)
{
  while (current != target && current != T_EOF)
    getNextToken();
  if (current == T_EOF) return 0;
  return 1;
}

int Tokeniser::expect(Token expected)
{
  Token token = getNextToken();
  if (token != expected)
    return error("Unexpected token '%s' where '%s' expected", tokenStrings[token], tokenStrings[expected]);
  return 1;
}

int Tokeniser::scanCodeBlock(CodeBlock &code, int indent)
{ 
  int scope = 1, length = 0, start = lineNumber, startLine = 1, s;
  if (!expect(T_OPEN_BRACE)) return 0;

  suppressComment = 1;
  code.setLength(128);

  for (s = 0; s < indent; s++, length++) {
    code.check(length);
    code[length] = ' ';
  }

  while (scope) {
    switch(ch) {
    case END_OF_FILE:
      error("Unclosed code block starting on line %d", start);
      return 0;
      break;
    case '\n':
      startLine = 1;
      if (length > indent) {
        code.check(length);
        code[length] = '\n';
        length++;
        for (s = 0; s < indent; s++, length++) {
          code.check(length);
          code[length] = ' ';
	}
      }
      break;
    case '{': 
      indent += 2;
      scope++;
      break;
    case '}':
      indent -= 2;
      if (indent < 0) indent = 0;
      scope--;
      break;
    }
    if (scope) {
      if (!startLine || !whitespace(ch)) {
        startLine = 0;
        code.check(length);
        code[length] = ch;
        length++;
      }
    }
    nextChar();
  }
  suppressComment = 0;
  while (length && whitespace(code[length - 1])) length--;
  code.check(length);
  code[length] = '\0';
  length++;
  code.setLength(length);
  #ifdef DEBUG
  cout << "Successfully read code block: " << endl;
  cout << code; // << "(" << (int) code[n] << ")\n";
  cout << endl << "(end)" << endl;
  #endif
  return 1;
}

char *Tokeniser::getTokenText(Token token) {
  for (int n=0; table[n].token != T_UNDEFINED ; n++)
    if (table[n].token == token)
      return table[n].keyword;
  return tokenStrings[token];
}

int Tokeniser::unexpected(Token token, char *context) {
  if (token == T_IDENTIFIER)
    return error("Unexpected token '%s' in %s", identifier, context);
  return error("Unexpected token '%s' in %s", getTokenText(token), context);
}

