

// Filename: uniform.cxx
// Author:   Nick Dingle & Will Knottenbelt {njd200,wjk}@doc.ic.ac.uk
// Date :    18th September 2002
// Notes:    Uniformisation-based response-time analyser


using namespace std;

#include <fstream>
#include <iostream>
#include <iomanip>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <ctime>
#include <cmath>
#include <complex>
#include <unistd.h>
#include <cstdarg>
#include <sys/stat.h>
#include <sys/types.h>
#include <cassert>
#include "common.hpp"
#include "dyna.hpp"
#include "timer.hpp"
#include "info.hpp"
#include "element.hpp"
#include "vector.hpp"
#include "store.hpp"
#include "avl.hpp"
#include "sparse-nocomplex.hpp"
#include "uniform.hpp"
#include "filenames.hpp"
#include "bit.hpp"
//#include "mpi.h"

#undef real

#define NO_DISK_LOG

#define MAX_N 10000

//#define CDF

#define BUFFER_SIZE 131072
#define TAG_COUNT  100
#define TAG_INDEX  200
#define TAG_RESULT 300
#define TAG_APPLY  400
#define TAG_WORK   500
#define TAG_EXACT  600
#define TAG_TVALUE 700
#define TAG_SVALUE 800

#define NO_EVALUATE 0
#define EVALUATE 1
#define USE_CACHE 2

//#include "value.hpp"
//#include "laplace.hpp"

// scaling factor for courier model
#define FACTOR 1.0

class Uniformiser {
  void *core;
  
  Information *info;
  Timer *timer;

  void destroy();

  long trans, tangible;

  int instance, size, targets;
  int cachehits, doCDF;
  int doPassage, doTransient;

  int *t_max_hops;

  double t_bottom, t_top, t_step;

  /* target state for analysis */
  long source_state;
  BitField target_states;

  FastDynaArray<double> diagonal, rhs, sum;
  Vector<double> weight;

  void terminate(char *fmt, ...);
  int calc_left_terms(double *terms, int max, double t);
  double log_erlang(int n, double q, double t);
  double log_erlang_cdf(int n, double q, double t);

  SparseMatrix *Q;
  double _q;

 
  inline void write_to_memory() {
    //    int **block = new (int*)[1024];
    typedef int* intp;
    int **block = new intp[1024];

    for(int i=0; i<1024; i++) {
      block[i] = new int[2];
    }
    for(int i=0; i<1024 * 2; i++) {
      int x = (int) (drand48() * 1024);
      int y = (int) (drand48() * 2);
      int z = (int) (drand48() * 10);
      block[x][y] = z;
      //	cout << ID_PASSAGE << "block[" << x << "][" << y << "]=" << block[x][y] << "    DEBUG" <<  endl;
    }
    //    cout << ID_PASSAGE << "finished write, deleting..." << endl;
    for(int i=0; i<1024; i++)
      delete [] block[i];
    delete [] block;
  }
 

public:
  Uniformiser(int _instance, int _size, char* basename, int _cdf);
  ~Uniformiser();

  int uniformise(char*);
  int readStates();

  int calculate(); 

  //  double exact(double T); 

};

void Uniformiser::terminate(char *fmt, ...)
{
  char buffer[2048];

  va_list argp;
  va_start(argp, fmt);
  vsprintf(buffer, fmt, argp); 
  va_end(argp);
  destroy();
  printf("\nfatal error: %s.\nexecution halted.\n", buffer);  
  exit(1);
}

double Uniformiser::log_erlang(int n, double q, double t) {
  if(doPassage) {
    long double logn = 0;
    for (int k=2; k<n; k++)
      logn += log((long double)k);
    logn += q*t;
    long double logp = (double) n*log(q)+ (double) (n-1)*log(t) - logn;
    return exp(logp);
  } else {
    long double logn = 0;
    for (int k=2; k<=n; k++)
      logn += log((long double)k);
    logn += q*t;
    long double logp = (double) n*log(q)+ (double) (n)*log(t) - logn;
    return exp(logp);
  }
}

double Uniformiser::log_erlang_cdf(int n, double q, double t) {

  long double sum = 0;

  long double logn=0;
  for(int i=0; i<n; i++) {
    //    long double logn = 0;
    //    for (int k=2; k<=i; k++)
    //      logn += log(k);
    if (i>=2)
      logn += log((long double)i);
    //    assert(fabs(logn2 - logn) < 1e-05);

    long double logc = i*(log(q) + log(t)) - (logn + q*t);
    sum += exp(logc);
  }
  if (sum > 1.0) sum = 1.0;
  return 1.0 - sum;
}

int Uniformiser::calc_left_terms(double *terms, int max, double t) { 
  int n;

  /*
  double *a = new double[max];

  a[0] = 1;
  for (n=1; n<max; n++) {
    a[n] = a[n-1]*_q*t/n; 
  }
  */

  double sum = 0, prev = -1, cdfterm;

  // Initialise cdfterm, this is mostly to get rid of the warning
  // about the possible use of it uninitialised
  cdfterm = log_erlang_cdf(1,_q, t) ;
  terms[0] = 0;

  int had_non_zero = 0;
  for (n=1; n<max+2; n++) {

    double term = log_erlang(n,_q,t); // _q*a[n-1]*exp(-_q*t);  /*_q*a[n-1]*exp(-_q*t)**/

    if(doCDF) {
      //#ifdef CDF
      cdfterm = log_erlang_cdf(n,_q,t);
      //    cout << ID_UNIFORM << "n=" << n << " _q=" << _q << " t=" << t << " cdfterm=" << cdfterm << "  CDF" << endl;
      //#endif
    }
   
    //    if(t == 46 && n > 600) {
    //      cout << ID_UNIFORM << "n=" << n << " sum=" << sum << " term=" << term << " t=" << t << "  SUUM" << endl;
      //      cout << ID_UNIFORM << "_q*a[n-1]=" << _q*a[n-1] << " exp(-_q*t)=" << exp(-_q*t) << " _q=" << _q << "  a[n-1]=" << a[n-1] << " SUUM" << endl;

      //   }
   

    sum += term;

    if(!doCDF)
      terms[n] = term;
    else
      terms[n] = cdfterm;

    /*
#ifndef CDF
    terms[n] = term;
#else
    terms[n] = cdfterm;
#endif
    */

    if (term > 1e-20)
      had_non_zero = 1;

    //    if (t>=70) 
    //    cout << "term t = " << t << " n = " << n << " term = " << term << " sum = " << sum << endl;

    if (n > 5 && term<1e-30 && had_non_zero) break;
    prev = sum;

#ifdef LOGGING
    if(!(n%100))
      cout << ID_UNIFORM << "term t = " << t << " n = " << n << endl;
#endif

    if(n==max) {
      cout << ID_UNIFORM << "erlang terms have not decayed to 0 by n=" << n << "..." << endl;
      cout << ID_UNIFORM << "suggest lowering the t-range or reducing rates..." << endl;
      exit(-1);
    }

  }


  //  delete [] a;
  
  return n;
}

Uniformiser::Uniformiser(int _instance, int _size, char* basename, int _cdf)
{
   char* optionsFileName = add_extension(basename, "OPTIONS") ;
   char* infoFileName    = add_extension(basename, "INFO") ;
   char* weightFileName  = add_extension(basename, "WEIGHT") ;
   char* targetsFileName = add_extension(basename, "TARGETS") ;

#ifdef LOGGING
  cout << ID_UNIFORM << "powering up..." << endl;
#endif

#ifdef RESOURCE_USAGE
  core = sbrk(0);
#endif

  //  ltc = new LaplaceTransformCalculator(_instance, _size);

  instance = _instance;
  size = _size;
  doCDF = _cdf;
  targets = 0;
  cachehits = 0;

  trans = 0;

  info = new Information;
  timer = new Timer;
  
  if (!info->input(optionsFileName)) 
     terminate("uniform.cxx cannot read options file");
  if (!info->inputStateInfo(infoFileName))
     terminate("uniform.cxx cannot read state information file");

  if (!info->getGeneration()) terminate ("run state space generator first");
  if (!info->getTangible()) terminate("no tangible states");
  tangible = info->getTangible();

  if (!info->getFunctional())
    cout << "warning: no functional analysis has been performed on model. recommend" << endl << "         running functional analyser before proceeding" << endl << endl;
  //SolutionMethod method = info->getMethod();

  doPassage = info->getPassage();
  doTransient = info->getTransient();

  if(!doPassage) {
    assert(doTransient);
    doCDF = 0;
#ifdef LOGGING
    cout << ID_UNIFORM << "performing transient analysis..." << endl;
#endif
  }
  else{
    assert(doPassage);
#ifdef LOGGING
    cout << ID_UNIFORM << "performing passage time analysis..." << endl;
#endif
  }


  Q = new SparseMatrix(tangible, tangible);
     
  t_bottom = info->getTStart();
  t_top = info->getTStop();
  t_step = info->getTStep();

  source_state = 0;
  target_states.setLength(tangible);
  target_states.clear();

  Vector<int> target;
  target.input(targetsFileName);

  //  if (instance == 0) 
  //    cout << "target states: [";

#ifdef LOGGING
  cout << ID_UNIFORM << "there are " << target.getLength() << " target states..." << endl;
#endif

  for (int n=0; n<target.getLength(); n++) {
    //    if (!instance && n) cout << " ";
    //    if (instance == 0) 
    //      cout << target[n];
    target_states.set1(target[n]);
  }

  //  if (instance == 0)
  //    cout << "]" << endl;

  if (target.getLength() == 0) 
    terminate("no target states found");

  weight.input( weightFileName );
  if (weight.getLength() != tangible)
    terminate("invalid initial state weight vector");
  
  //  write_log("[%d] powered up!", instance);

#ifdef LOGGING
  cout << ID_UNIFORM << "powered up!" << endl;
#endif
}

Uniformiser::~Uniformiser() {
  destroy();
}

void Uniformiser::destroy() {
#ifdef LOGGING
  cout << ID_UNIFORM << "destructor..." << endl;
#endif
  delete info;
  delete timer;
  if(t_max_hops != NULL) {
    //    cout << ID_UNIFORM << "destroying t_max_hops..." << endl;
    delete [] t_max_hops;
  }
#ifdef LOGGING
  cout << ID_UNIFORM << "powered down..." << endl;
#endif
}

int calls = 0;

int Uniformiser::readStates() {
#ifdef LOGGING
  cout << ID_UNIFORM << "reading tangible states..." << endl;
#endif
  timer->start();

  diagonal.setLength(tangible);
  rhs.setLength(tangible);
  sum.setLength(tangible);

  for (int n=0; n<tangible; n++)
    rhs[n] = sum[n] = 0;

  char *filename = info->getTransMatrixFile();
  
  char buffer[BUFFER_SIZE];
  FILE *mf = fopen(filename,"rb");
  if (!mf) terminate("cannot open state matrix file '%s'", filename);
  setvbuf(mf, buffer, _IOFBF, BUFFER_SIZE);

  int kids, states = 0;
  StateMatrixElement *child = new StateMatrixElement[info->getMaxChildren()];

  _q = 0;
  while(fread(&kids, sizeof(int), 1, mf)) {
    fread(child, sizeof(StateMatrixElement), kids, mf); 
    double _sum = 0;
    for (int t=0; t<kids; t++) {
      //      register double _dest = child[t].getDestination();
      register long _dest = child[t].getDestination();
      register double _weight = child[t].getWeight();
      if (_dest != states) 
	_sum += _weight;
    }
    if (_sum > _q)
      _q = _sum;
    states++;
    if (!(states % 10000) && (instance == 0))
      cout << ID_UNIFORM << "reading state " << states << "..." << endl;
  }
#ifdef LOGGING
  cout << ID_UNIFORM << "_q = " << _q << endl;
#endif
  _q *= 1.01;

#ifdef LOGGING
  cout << ID_UNIFORM << "modified _q = " << _q << endl;
#endif

  assert(states == tangible);
  fclose(mf); 

  mf = fopen(filename,"rb");
  if (!mf) terminate("cannot open state matrix file '%s'", filename);
  setvbuf(mf, buffer, _IOFBF, BUFFER_SIZE);

  states = 0;
  while(fread(&kids, sizeof(int), 1, mf)) {
    fread(child, sizeof(StateMatrixElement), kids, mf); 
    
    double _sum = 0;
    rhs[states] = 0;

    for (int t=0; t<kids; t++)  {
      register long _dest = child[t].getDestination();
      register double _weight = child[t].getWeight();

      if(doPassage) {
	if (target_states(states) && _dest != states) {
	  _sum += _weight;
	  //	rhs[states] += _weight;
	} else 
	  if (_dest != states) {
	    _sum += _weight;
	    Q->addElement(states, _dest, _weight/_q);
	  }
      }
      else{
	if (_dest != states) {
	  _sum += _weight;
	  Q->addElement(states, _dest, _weight/_q);
	}
      }
    }


    if(doPassage) {
      //    diagonal[states] = _sum/_q + 1.0;
      if (!target_states(states))
	Q->addElement(states, states, 1.0 - _sum/_q);
      //    else
      //      Q->addElement(states, tangible, 1.0);
      
      //    Q->getRow(states).lock();
    } else {
      Q->addElement(states, states, 1.0 - _sum/_q);
    }

    //    Q->getRow(states).lock();

    trans += kids;
    states++;

    //Q->addElement(tangible,tangible,1.0);

    if (!(states % 10000) && (instance == 0))
      cout << ID_UNIFORM << "reading state " << states << "... (pass 2)" << endl;
  }

  assert(states == tangible);
  delete [] child;
  fclose(mf); 

#ifdef LOGGING
  cout << ID_UNIFORM << "read complete ( " << tangible << " states, " << trans << " transitions) " << endl;
#endif

  double timeUsed = timer->read();
  if (timeUsed > 0) {
    cout << ID_UNIFORM << "(read speed " << ((double) tangible/timeUsed) << " states/second, " << ((double) trans/timeUsed) << " transitions/second)" << endl;
  }

  return 1;
}

int Uniformiser::uniformise(char* basename) {
   char* ptResultsFileName = add_extension(basename, "PT_RESULTS") ;

#ifdef LOGGING
  cout << ID_UNIFORM << "ready (" << tangible << " states)!" << endl;
#endif

  readStates();

  Vector<double> pi, result;

  pi.setLength(tangible);
  result.setLength(tangible);

  //  int no_of_t_points = (int) (t_top/t_step);

  //  cout << "t_top/t_step=" << (int)(t_top/t_step) << " MAX_N=" << MAX_N << endl;

  //  double terms[(int) (t_top/t_step)][MAX_N];

  double **terms = new double*[(int) (t_top/t_step)];

  for(int i=0; i<(int) (t_top/t_step); i++) 
    terms[i] = new double[MAX_N];

  //  cout << "setup terms" << endl;

  FastDynaArray<double> sum_pi_target;
  //  int t_max_hops[(int) (t_top/t_step)];

  //  int *t_max_hops = new int[(int) (t_top/t_step)];
  t_max_hops = new int[(int) (t_top/t_step)];
  //  int *t_max_hops = new int[no_of_t_points];
  assert(t_max_hops);

  //  cout << "setup t_max_hops" << endl;

  int max_hops = -1;

#ifdef LOGGING
  cout << ID_UNIFORM << "preparing to assign values to t_max_hops and lhs..." << endl;
#endif

  for (int count=0; count < (int)(t_top/t_step); count++) {
       //  for (double t = T_BOTTOM; t < T_TOP+T_STEP; t += T_STEP) {
    t_max_hops[count] = calc_left_terms(terms[count], MAX_N, (count+1)*t_step);

    //    cout << ID_UNIFORM << "max_hops[" << count << "] = " << t_max_hops[count] << endl;

    if (t_max_hops[count] > max_hops) 
      max_hops = t_max_hops[count];
  

  //      for (int n=0; n<(T_TOP/T_STEP); n++)
    //	t_max_hops[count] = calc_left_terms(terms[n], 1000, (count+1)*T_STEP);
  }

#ifdef LOGGING
  cout << ID_UNIFORM << "done" << endl;
#endif

  /*
  cout << "Calculating value for max_hops..." << endl;

  for(int i=0; i<(T_TOP/T_STEP); i++) {
    cout << t_max_hops[i] << " ";
      if (t_max_hops[i] > max_hops) 
	max_hops = t_max_hops[i];
	}
  cout << endl;

  cout << "Done" << endl;
  */

#ifdef LOGGING
  cout << ID_UNIFORM << "max_hops is " << max_hops << endl;

  cout << ID_UNIFORM << "setting length of sum_pi_target to " << (max_hops + 1) << endl;
#endif

  sum_pi_target.setLength(max_hops + 1);
  //  cout << "Done" << endl;

#ifdef LOGGING
 cout << ID_UNIFORM << "setting up pi_0..." << endl;
#endif

  for (int n=0; n<tangible; n++)
    pi[n] = weight[n];

#ifdef LOGGING
  //  double start_time = timer->read();
  double start_time = timer->observe();
#endif

  int converged = 0;
  for (int hop = 1; hop <= max_hops; hop ++) {

    // cout << "hop=" << hop << ". Done" << endl;  

    if (!converged)
      Q->transMultiply(pi, result);

    sum_pi_target[hop] = 0;

    converged = 1;
    for (int n=0; n<tangible; n++) {
      if (fabs(pi[n]-result[n]) > 1e-10)
        converged = 0;
      pi[n] = result[n];
      if (target_states(n)) {
	sum_pi_target[hop] += pi[n];
      }
    }

    // cout << ID_UNIFORM << "sum_pi_target[" << hop << "]=" << sum_pi_target[hop] << endl; 

    //    double elap = timer->observe()-st;

    if(!(hop%100)) {
#ifdef LOGGING
      cout << ID_UNIFORM << "done hop " << hop << " of " << max_hops << " ..." << endl;
#endif
      write_to_memory();
    }
  }

  write_to_memory();


#ifdef LOGGING
  //  double stop_time = timer->read();
  double stop_time = timer->observe();
  double elapsed = stop_time - start_time;
  //  cout << ID_UNIFORM << "start_time=" << start_time << " stop_time=" << stop_time << endl;

  cout << ID_UNIFORM << "done matrix multiplication (pi calculation)..." << endl;
#endif


    /*    int max_hops = calc_left_terms(terms, 1000, t);

	  sum_pi_target.setLength(max_hops);*/

    /*double sum = 0;
    cout << "hop = " << 0 << " pi = [";
    for (int n=0; n<tangible; n++) {
      if (n) cout << " ";
      cout << pi[n];
      sum += pi[n];
    } 
    cout << "] (sum = " << sum << ")" << endl;*/

    /*   for (int hop = 1; hop < max_hops; hop ++) {

      Q->transMultiply(pi, result);

      sum_pi_target[hop] = 0;
      for (int n=0; n<tangible; n++) {
	pi[n] = result[n];
	if (target_states(n))
	  sum_pi_target[hop] += pi[n];
	  }*/

      /* double sum = 0;
      cout << "hop = " << hop << " pi = [";
      for (int n=0; n<tangible; n++) {
	if (n) cout << " ";
	cout << pi[n];
	sum += pi[n];
      } 
      cout << "] (sum = " << sum << ")" << endl;*/

    //    cout << "sum_pi_target[hop = " << hop << "] = " << sum_pi_target[hop] << endl;
    //    }

    //  }


#ifdef LOGGING
  cout << ID_UNIFORM << "calculating answers for each t_point..." << endl;
#endif

 // for (int hop = 1; hop <= max_hops; hop++) {
 //   cout << "HOP " << hop << " " << sum_pi_target[hop] << endl;
 // }

  //  FILE *result_output = fopen("PT_RESULTS", "wt");
  //  assert(result_output);

  ofstream result_output;
  result_output.open(ptResultsFileName);

  for (int count=0; count < (int)(t_top/t_step); count++) {

    double answer = 0.0;

    for (int hop = 1; hop <= t_max_hops[count]; hop++) {
      answer += terms[count][hop] * sum_pi_target[hop];
    }

#ifdef LOGGING
    cout << (count+1)*t_step << "    " << answer <<  "    DATA0" << endl;
#endif

    result_output << (count+1)*t_step << "    " << answer << endl;

    //    fprintf(result_output, "%f  %f10\n", (count+1)*t_step, answer);

    //    write_log("%g  %g  DATA0", (count+1)*t_step, answer);

   }

#ifdef LOGGING
  cout << "# elapsed matrix multiply time was " << elapsed << " seconds for " << max_hops << " iterations   DATA0" << endl;
#endif
  //  write_log("[%d] elapsed time was %g for %d iterations  DATA0", instance, elapsed, max_hops);

  //  fclose(result_output);
  result_output.close();

  for(int i=0; i<(int) (t_top/t_step); i++) 
    delete [] terms[i];

  delete [] terms;

  //  cout << ID_UNIFORM << "terms deleted successfully..." << endl;

  // if we try to delete [] t_max_hops, we get a segfault when number of t-points is large - why?
  // fixed -  made it a class variable & destroyed it in destroy() 
  //  delete [] t_max_hops;
  //  cout << ID_UNIFORM << "t_max_hops deleted successfully..." << endl;

  return 0;
}


void change_path(char *fullname) {
  char *path = new char[strlen(fullname) + 1];
  strcpy(path, fullname);
  char *target = strrchr(path, '/');
  if (target) {
    *target = '\0';
    cout << "directory = " << path << endl;
    chdir(path);
  }    
  delete [] path;
}


int main(int argc, char *argv[]) {

  int instance, size, cdf=0;
  //double start, end;
  //  MPI_Init(&argc, &argv);
  //  MPI_Comm_size(MPI_COMM_WORLD, &size);
  //  MPI_Comm_rank(MPI_COMM_WORLD, &instance);

  //change_path(argv[0]);

  instance = 0;
  size = 1;

  if(argc==3)
    cdf = 1;
#ifdef LOGGING
  cout << "Initialising Uniformiser..." << endl;
#endif
  Uniformiser *alex = new Uniformiser(instance, size, argv[1], cdf);

#ifdef LOGGING
  cout <<"Running uniformise();" << endl;
#endif
  alex->uniformise(argv[1]);

  delete alex;

  //  MPI_Finalize();

  return 0;
}
