////////////////////////////////////////////////////////////////////
// WMCA Markov Chain Analyser        (C) 1995 William Knottenbelt //
////////////////////////////////////////////////////////////////////
//            Data Network Architectures Laboratory               //
//               Department of Computer Science                   //
//                   University of Cape Town                      //
//                       Rondebosch 7700                          //
//                        SOUTH AFRICA                            //
////////////////////////////////////////////////////////////////////
// email: william@cs.uct.ac.za             phone: +27 21 650 2670 //
////////////////////////////////////////////////////////////////////

#ifndef __VECTOR_CXX
#define __VECTOR_CXX

using namespace std;

#include <fstream>
#include <iomanip>
#include <cstdio>
#include <cmath>
#include "vector.hpp"

#ifndef LOGGING
#define SILENT 1
#endif

template <class T>
Vector<T>::Vector() : FastDynaArray<T>()
{
}

template <class T>
Vector<T>::Vector(long size) : FastDynaArray<T>(size)
{
}

template <class T>
Vector<T>::~Vector()
{
}

template <class T>
Vector<T>::Vector(const Vector<T> &source)
{
  #ifndef SILENT
  cout << "Vector copy constructor (" << this << ")\n";
  #endif
  FastDynaArray<T>::setLength(source.FastDynaArray<T>::length);
  for (long n=0; n<FastDynaArray<T>::length; n++)
    FastDynaArray<T>::array[n] = source(n);
}

template <class T>
T Vector<T>::oneNorm()
{
  #ifdef VERBOSE
  cout << "Vector (" << this << "): finding oneNorm\n";
  #endif

  T sum = 0;

  for (long n=0; n<FastDynaArray<T>::length; n++) {
    if (FastDynaArray<T>::array[n] < 0)
      sum -= FastDynaArray<T>::array[n];
    else
      sum += FastDynaArray<T>::array[n];
  }
  return sum;
}

template <class T>
T Vector<T>::twoNorm()
{
  #ifdef VERBOSE
  cout << "Vector (" << this << "): finding twoNorm\n";
  #endif

  T sum = 0;

  for (long n=0; n<FastDynaArray<T>::length; n++)
    sum += FastDynaArray<T>::array[n]*FastDynaArray<T>::array[n];

  return (T) sqrt((double)sum);
}

template <class T>
T Vector<T>::infNorm()
{
  #ifdef VERBOSE
  cout << "Vector (" << this << "): finding infNorm\n";
  #endif

  T max = 0;

  for (long n=0; n<FastDynaArray<T>::length; n++) {
    if (FastDynaArray<T>::array[n] < 0) {
      if (-FastDynaArray<T>::array[n]>max) max = -FastDynaArray<T>::array[n];
    }
    else
      if (FastDynaArray<T>::array[n]>max) max = FastDynaArray<T>::array[n];
  }

  return max;
}

template <class T>
int Vector<T>::normalise()
{
  #ifdef VERBOSE
  cout << "Vector (" << this << "): normalising\n";
  #endif

  return divide(oneNorm());
}

template <class T>
int Vector<T>::normalise(long n)
{
  #ifdef VERBOSE
  cout << "Vector (" << this << "): normalising\n";
  #endif

  return divide(oneNorm()/n);
}

template <class T>
Vector<T> &Vector<T>::operator=(const Vector<T> &source)
{
  #ifdef VERBOSE
  cout << "Vector operator = (" << this << ")\n";
  #endif
  if (this == &source) return (*this);
  FastDynaArray<T>::setLength(source.FastDynaArray<T>::length);
  for (long n=0; n<FastDynaArray<T>::length; n++)
    FastDynaArray<T>::array[n] = source(n);
  return (*this);
}

template <class T>
Vector<T> Vector<T>::operator-(const Vector<T> &source) const
{
  #ifdef VERBOSE
  cout << "Vector operator -\n";
  #endif

  Vector<T> result(source.FastDynaArray<T>::length);

  #ifdef SAFE
  if (length != source.FastDynaArray<T>::length) {
    cout << "Vector operator -: Warning! vectors are incompatible\n";
    return result;
  }
  #endif
 
  for (long n=0; n<FastDynaArray<T>::length; n++)
    result[n] = FastDynaArray<T>::array[n] - source(n);

  return result;
}

template <class T>
Vector<T> Vector<T>::operator+(const Vector<T> &source) const
{
  #ifdef VERBOSE
  cout << "Vector operator +\n";
  #endif

  Vector<T> result(source.FastDynaArray<T>::length);

  #ifdef SAFE
  if (length != source.FastDynaArray<T>::length) {
    cout << "Vector operator +: Warning! vectors are incompatible\n";
    return result;
  }
  #endif

  for (long n=0; n<FastDynaArray<T>::length; n++)
      result[n] = FastDynaArray<T>::array[n] + source(n);
  return result;
}

template <class T>
int Vector<T>::multiply(T factor)
{
  #ifdef VERBOSE
  cout << "Vector ("<< this << ") multiply (" << factor << ")\n";
  #endif

  for (long n=0; n<FastDynaArray<T>::length; n++)
    FastDynaArray<T>::array[n] *= factor;
  return 1;
}

template <class T>
int Vector<T>::divide(T factor)
{
  #ifdef VERBOSE
  cout << "Vector ("<< this << ") divide (" << factor << ")\n";
  #endif

  if (factor == 0) {
    #ifndef SILENT
    cout << "Vector divide: Warning! Attempted division by zero\n";
    #endif
    return 0;
  }
  if (factor == 1.0) return 1;
  T scale = (T) 1.0/factor;
  for (long n=0; n<FastDynaArray<T>::length; n++)
    FastDynaArray<T>::array[n] *= scale;
  return 1;
}

template <class T>
int Vector<T>::inputError(const char *filename)
{
  #ifndef SILENT
  cout << "Vector: Warning! Input failure (reading '" << filename << "')\n";
  #endif
  return 0;
}

template <class T>
int Vector<T>::outputError(const char *filename)
{
  #ifndef SILENT
  cout << "Vector: Warning! Output failure (writing '" << filename << "')\n";
  #endif
  return 0;
}

template <class T>
int Vector<T>::input(const char *filename)
{
  #ifndef SILENT
  cout << "Vector input (" << this << ") from '" << filename << "'\n";
  #endif
  long inLength;

  FILE *vf = fopen(filename,"rb");
  if (!vf) return inputError(filename);

  inLength = -1;
  int success = fread(&inLength, sizeof(long), 1, vf);
  if (!success || inLength < 0) return inputError(filename);

  FastDynaArray<T>::setLength(inLength);
  success = fread(FastDynaArray<T>::array, sizeof(T), FastDynaArray<T>::length, vf);

  if (!success) return inputError(filename);
  fclose(vf);
  return 1;
}

template <class T>
int Vector<T>::textInput(const char *filename)
{
  #ifndef SILENT
  cout << "Vector text input (" << this << ") from '" << filename << "'\n";
  #endif
  long inLength;
  ifstream vect(filename, ios::in);
  if (!vect) return inputError(filename);

  vect >> inLength;
  if (vect.fail() || inLength < 0) return inputError(filename);

  FastDynaArray<T>::setLength(inLength);
  for (long n=0; n<inLength; n++)
    vect >> FastDynaArray<T>::array[n];

  if (vect.fail()) return inputError(filename);
  vect.close();

  return 1;
}

template <class T>
int Vector<T>::output(const char *filename)
{
  #ifndef SILENT
  cout << "Vector output (" << this << ") to '" << filename << "'\n";
  #endif

  FILE *vf = fopen(filename, "wb");
  if (!vf) return outputError(filename);

  int success1, success2;
  long temp_length = FastDynaArray<T>::length;

  //  success1 = fwrite(&length, sizeof(long), 1, vf);

  success1 = fwrite(&temp_length, sizeof(long), 1, vf);
  success2 = fwrite(FastDynaArray<T>::array, sizeof(T), FastDynaArray<T>::length, vf);

  if (!success1 || !success2) return outputError(filename);

  fclose(vf);
  return 1;
}

template <class T>
int Vector<T>::textOutput(const char *filename, int precision)
{
  #ifndef SILENT
  cout << "Vector text output (" << this << ") to '" << filename << "'\n";
  #endif
  ofstream vect(filename, ios::out);
  if (!vect) return outputError(filename);

  vect << FastDynaArray<T>::length << setprecision(precision) << endl;
  if (vect.fail()) return outputError(filename);

  for (long n=0; n<FastDynaArray<T>::length; n++)
    vect << FastDynaArray<T>::array[n] << endl;

  if (!vect.good()) return outputError(filename);
  vect.close();

  return 1;
}

#endif


